from boat_jit.utils.op_utils import (
    grad_unused_zero,
    update_tensor_grads,
    copy_parameter_from_list,
    manual_update,
)
import numpy
from jittor import Module
import copy
from typing import Dict, Any, Callable, List
import jittor as jit

from boat_jit.operation_registry import register_class
from boat_jit.dynamic_ol.dynamical_system import DynamicalSystem


@register_class
class MESM(DynamicalSystem):
    """
    Implements the optimization procedure of Moreau Envelope based Single-loop Method (MESM) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective of the BLO problem.
    ul_objective : Callable
        The upper-level objective of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        The list of lower-level variables of the BLO problem.
    ul_var : List[jittor.Var]
        The list of upper-level variables of the BLO problem.
    lower_loop : int
        Number of iterations for lower-level optimization.
    solver_config : Dict[str, Any]
        A dictionary containing solver configurations. Expected keys include:

        - "lower_level_opt": The optimizer for the lower-level model.
        - "MESM" (Dict): A dictionary containing the following keys:
            - "eta": Learning rate for the MESM optimization procedure.
            - "gamma_1": Regularization parameter for the MESM algorithm.
            - "c0": Initial constant for the update steps.
            - "y_hat_lr": Learning rate for optimizing the surrogate variable `y_hat`.

    References
    ----------
    [1] Liu R, Liu Z, Yao W, et al. "Moreau Envelope for Nonconvex Bi-Level Optimization: A Single-loop and Hessian-free Solution Strategy," ICML, 2024.
    """

    def __init__(
        self,
        ll_objective: Callable,
        lower_loop: int,
        ul_model: Module,
        ul_objective: Callable,
        ll_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict[str, Any],
    ):
        super(MESM, self).__init__(
            ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
        )
        self.ll_opt = solver_config["lower_level_opt"]
        self.ll_var = ll_var
        self.ul_var = ul_var
        self.y_loop = lower_loop
        self.eta = solver_config["MESM"]["eta"]
        self.gamma_1 = solver_config["MESM"]["gamma_1"]
        self.c0 = solver_config["MESM"]["c0"]
        self.y_hat = copy.deepcopy(self.ll_model)
        self.y_hat_opt = jit.optim.SGD(
            self.y_hat.parameters(), lr=solver_config["MESM"]["y_hat_lr"], momentum=0.9
        )

    def optimize(self, ll_feed_dict: Dict, ul_feed_dict: Dict, current_iter: int):
        """
        Execute the optimization procedure with the data from feed_dict.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            It typically includes training data, targets, and other information required to compute the LL objective.
        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            It typically includes validation data, targets, and other information required to compute the UL objective.
        current_iter : int
            The current iteration number of the optimization process.

        Returns
        -------
        Dict
            A dictionary containing the upper-level objective and the status of hypergradient computation.
        """

        if current_iter == 0:
            ck = 0.2
        else:
            ck = numpy.power(current_iter + 1, 0.25) * self.c0

        theta_loss = self.ll_objective(ll_feed_dict, self.ul_model, self.y_hat)

        grad_theta_parmaters = grad_unused_zero(
            theta_loss, list(self.y_hat.parameters())
        )

        errs = []
        for a, b in zip(
            list(self.y_hat.parameters()), list(self.ll_model.parameters())
        ):
            diff = a - b
            errs.append(diff)
        vs_param = []
        for v0, gt, err in zip(
            list(self.y_hat.parameters()), grad_theta_parmaters, errs
        ):
            vs_param.append(v0 - self.eta * (gt + self.gamma_1 * err))  # upate \theta

        copy_parameter_from_list(self.y_hat, vs_param)

        reg = 0
        for param1, param2 in zip(list(self.ll_model.parameters()), vs_param):
            diff = param1 - param2
            reg += (diff**2).sum()  # Jittor-compatible L2 norm calculation
        lower_loss = (
            (1 / ck) * self.ul_objective(ul_feed_dict, self.ul_model, self.ll_model)
            + self.ll_objective(ll_feed_dict, self.ul_model, self.ll_model)
            - 0.5 * self.gamma_1 * reg
        )

        grad_y_parmaters = grad_unused_zero(
            lower_loss, list(self.ll_model.parameters())
        )

        update_tensor_grads(self.ll_var, grad_y_parmaters)

        manual_update(self.ll_opt, self.ll_var)
        upper_loss = (
            (1 / ck) * self.ul_objective(ul_feed_dict, self.ul_model, self.ll_model)
            + self.ll_objective(ll_feed_dict, self.ul_model, self.ll_model)
            - self.ll_objective(ll_feed_dict, self.ul_model, self.y_hat)
        )
        grad_x_parmaters = grad_unused_zero(
            upper_loss, list(self.ul_model.parameters())
        )
        update_tensor_grads(self.ul_var, grad_x_parmaters)

        return {"upper_loss": upper_loss.item()}



from boat_jit.utils.op_utils import l2_reg
from boat_jit.utils.op_utils import (
    update_grads,
    update_tensor_grads,
    grad_unused_zero,
    manual_update,
)
import jittor as jit
from jittor import Module
import copy
from typing import Dict, Any, Callable, List

from boat_jit.operation_registry import register_class
from boat_jit.dynamic_ol.dynamical_system import DynamicalSystem


@register_class
class VSM(DynamicalSystem):
    """
    Implements the optimization procedure of Value-function based Sequential Method (VSM) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        A list of lower-level variables of the BLO problem.
    ul_var : List[jittor.Var]
        A list of upper-level variables of the BLO problem.
    lower_loop : int
        The number of iterations for lower-level optimization.
    solver_config : Dict[str, Any]
        A dictionary containing configurations for the solver. Expected keys include:

        - "lower_level_opt" (torch.optim.Optimizer): Optimizer for the lower-level model.
        - "VSM" (Dict): Configuration for the VSM algorithm:
            - "z_loop" (int): Number of iterations for optimizing the auxiliary variable `z`.
            - "ll_l2_reg" (float): L2 regularization coefficient for the lower-level model.
            - "ul_l2_reg" (float): L2 regularization coefficient for the upper-level model.
            - "ul_ln_reg" (float): Logarithmic regularization coefficient for the upper-level model.
            - "reg_decay" (float): Decay rate for the regularization coefficients.
            - "z_lr" (float): Learning rate for optimizing the auxiliary variable `z`.
        - "device" (str): Device on which computations are performed, e.g., "cpu" or "cuda".

    References
    ----------
    [1] Liu B, Ye M, Wright S, et al. "BOME! Bilevel Optimization Made Easy: A Simple First-Order Approach,"
        in NeurIPS, 2022.
    """

    def __init__(
        self,
        ll_objective: Callable,
        lower_loop: int,
        ul_model: Module,
        ul_objective: Callable,
        ll_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict[str, Any],
    ):
        super(VSM, self).__init__(
            ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
        )
        self.ll_opt = solver_config["lower_level_opt"]
        self.ll_var = ll_var
        self.ul_var = ul_var
        self.y_loop = lower_loop
        self.z_loop = solver_config["VSM"]["z_loop"]
        self.ll_l2_reg = solver_config["VSM"]["ll_l2_reg"]
        self.ul_l2_reg = solver_config["VSM"]["ul_l2_reg"]
        self.ul_ln_reg = solver_config["VSM"]["ul_ln_reg"]
        self.reg_decay = float(solver_config["VSM"]["reg_decay"])
        self.z_lr = solver_config["VSM"]["z_lr"]

    def optimize(self, ll_feed_dict: Dict, ul_feed_dict: Dict, current_iter: int):
        """
        Execute the optimization procedure with the data from feed_dict.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            It typically includes training data, targets, and other information required to compute the LL objective.
        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            It typically includes validation data, targets, and other information required to compute the UL objective.
        current_iter : int
            The current iteration number of the optimization process.

        Returns
        -------
        Dict
            A dictionary containing the upper-level objective and the status of hypergradient computation.
        """

        reg_decay = self.reg_decay * current_iter + 1
        for z_idx in range(self.z_loop):
            self.ll_opt.zero_grad()
            loss_l2_z = self.ll_l2_reg / reg_decay * l2_reg(self.ll_model.parameters())
            loss_z_ = self.ll_objective(ll_feed_dict, self.ul_model, self.ll_model)
            loss_z = loss_z_ + loss_l2_z
            grads = grad_unused_zero(loss_z, list(self.ll_model.parameters()))
            update_grads(grads, self.ll_model)
            manual_update(self.ll_opt, list(self.ll_model.parameters()))

        auxiliary_model = copy.deepcopy(self.ll_model)
        auxiliary_opt = jit.nn.SGD(auxiliary_model.parameters(), lr=self.z_lr)

        with jit.no_grad():
            loss_l2_z = self.ll_l2_reg / reg_decay * l2_reg(self.ll_model.parameters())
            loss_z_ = self.ll_objective(ll_feed_dict, self.ul_model, self.ll_model)
            loss_z = loss_z_ + loss_l2_z

        for y_idx in range(self.y_loop):
            loss_y_f_ = self.ll_objective(ll_feed_dict, self.ul_model, auxiliary_model)
            loss_y_ = self.ul_objective(ul_feed_dict, self.ul_model, auxiliary_model)
            loss_l2_y = l2_reg(auxiliary_model.parameters())
            loss_l2_y = self.ul_l2_reg / reg_decay * loss_l2_y
            loss_ln = jit.log(loss_y_f_.item() + loss_z.item() - loss_y_f_.item())
            loss_ln = self.ul_ln_reg / reg_decay * loss_ln
            loss_y = loss_y_ - loss_ln + loss_l2_y
            grads = jit.grad(loss_y, auxiliary_model.parameters())
            update_grads(grads, auxiliary_model)
            manual_update(auxiliary_opt, list(auxiliary_model.parameters()))

        with jit.no_grad():
            loss_l2_z = self.ll_l2_reg / reg_decay * l2_reg(self.ll_model.parameters())
            loss_z_ = self.ll_objective(ll_feed_dict, self.ul_model, self.ll_model)
            loss_z = loss_z_ + loss_l2_z

            loss_y_f_ = self.ll_objective(ll_feed_dict, self.ul_model, auxiliary_model)
            loss_ln = (
                self.ul_ln_reg
                / reg_decay
                * jit.log(loss_y_f_.item() + loss_z.item() - loss_y_f_.item())
            )

        loss_x_ = self.ul_objective(ul_feed_dict, self.ul_model, auxiliary_model)
        loss_x = loss_x_ - loss_ln
        grads = jit.grad(loss_x, self.ul_model.parameters())
        update_tensor_grads(self.ul_var, grads)
        return {"upper_loss": loss_x_.item()}



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Differentiable optimizer wrappers around ``jittor.optim`` instances."""

import abc as _abc
import collections as _collections
import copy as _copy
import math as _math
import typing as _typing
import warnings as _warnings

import jittor as jit

from . import patch as _patch
from . import utils as _utils

_GroupedGradsType = _typing.List[_typing.List[jit.Var]]
_StateType = _typing.List[_typing.DefaultDict[int, _typing.Any]]
_GradClosureType = _typing.Callable[[jit.Var], jit.Var]
_OverrideType = _typing.Dict[str, _typing.List[_typing.Any]]
_GradCallbackType = _typing.Callable[[_typing.List[jit.Var]], _typing.List[jit.Var]]


def _get_mask_closure(mask: jit.Var) -> _GradClosureType:
    def closure(grad: jit.Var) -> jit.Var:
        grad = jit.where(mask, jit.zeros_like(grad), grad)
        if grad.requires_grad:
            grad.register_hook(_get_mask_closure(mask))
        return grad

    return closure


def _maybe_mask(tensor: jit.Var, mask: jit.Var) -> None:
    if tensor.requires_grad:
        tensor.register_hook(_get_mask_closure(mask))


class DifferentiableOptimizer(_abc.ABC):
    def __init__(
        self,
        other: jit.optim.Optimizer,
        reference_params: _typing.Iterable[jit.Var],
        fmodel: _typing.Optional[_patch._MonkeyPatchBase] = None,
        device: _typing.Optional[str] = None,
        override: _typing.Optional[_OverrideType] = None,
        grad_callback: _typing.Optional[_GradCallbackType] = None,
        track_higher_grads: bool = True,
        **kwargs,
    ) -> None:
        r"""Initialize the optimizer with the state of an existing optimizer.

        Args:
            other: an existing optimizer instance.
            reference_params: an iterable over the parameters of the original
                model.
            fmodel (optional): a patched stateless module with a view on
                weights.
            device (optional): the device to cast state tensors to.
            override (optional): a dictionary mapping optimizer settings (i.e.
                those which would be passed to the optimizer constructor or
                provided within parameter groups) to either singleton lists of
                override values, or to a list of override values of length equal
                to the number of parameter groups. If a single override is
                provided for a keyword, it is used for all parameter groups. If
                a list is provided, the ``i``\ th element of the list overrides the
                corresponding setting in the ``i``\ th parameter group. This permits
                the passing of tensors requiring gradient to differentiable
                optimizers for use as optimizer settings.
            grad_callback: (optional) a single argument function which will be
                applied to a list of gradients of parameters, which respects the
                order specified by ``reference_params``. This can be used to
                apply a function, such as gradient clipping, to all (or a
                subset) of these gradients every time the step function is
                called. If this keyword argument is provided when calling the
                step method, its value will override the default specified here.
            track_higher_grads: if True, during unrolled optimization the graph
                be retained, and the fast weights will bear grad funcs, so as to
                permit backpropagation through the optimization process. Setting
                this to False allows the differentiable optimizer to be used in
                "test mode", without potentially tracking higher order
                gradients. This can be useful when running the training loop at
                test time, e.g. in k-shot learning experiments, without
                incurring a significant memory overhead.
        """
        reference_params = list(reference_params)

        # Copy param groups and set up structures for copy state
        self.lr = other.lr
        self.param_groups = _copy.deepcopy(other.param_groups)
        self._group_to_param_list: _typing.List[_typing.List[int]] = []

        self.state: _typing.List[_typing.Dict[int, _typing.Dict[str, _typing.Any]]] = [
            _collections.defaultdict(dict) for _ in range(len(self.param_groups))
        ]

        # Deal with override
        if override is not None:
            self._apply_override(override)

        self._grad_callback = grad_callback

        # Initialize state
        zipped = zip(self.param_groups, other.param_groups)
        for group_idx, (group, orig_group) in enumerate(zipped):
            local_list = []
            for param_idx, param in enumerate(orig_group["params"]):
                param_id = id(param)  # Use the unique ID of the parameter
                if param_id in other._grad_map:
                    # Initialize state for the parameter
                    self.state[group_idx][param_idx] = {
                        k: _utils._recursive_copy_and_cast(v, device)
                        for k, v in other._grad_map[param_id].items()
                    }
                index = _utils._find_param_in_list(param, reference_params)
                if index is None:
                    raise ValueError(
                        f"Could not find parameter {param} in reference parameters."
                    )
                local_list.append(index)
            self._group_to_param_list.append(local_list)

        self._fmodel = fmodel
        self._track_higher_grads = track_higher_grads

    def _apply_override(self, override: _OverrideType) -> None:
        for k, v in override.items():
            # Sanity check
            if (len(v) != 1) and (len(v) != len(self.param_groups)):
                raise ValueError(
                    "Mismatch between the number of override tensors for "
                    "optimizer parameter {} and the number of "
                    "parameter groups.".format(k)
                )
            for group_idx, group in enumerate(self.param_groups):
                group[k] = v[0] if len(v) == 1 else v[group_idx]

    def step(
        self,
        loss: jit.Var,
        params: _typing.Iterable[jit.Var] = None,
        override: _typing.Optional[_OverrideType] = None,
        grad_callback: _typing.Optional[_GradCallbackType] = None,
        **kwargs,
    ) -> _typing.Iterable[jit.Var]:
        r"""Perform a model update.

        This would be used by replacing the normal sequence::

            opt.zero_grad()
            loss.backward()
            opt.step()

        with::

            diffopt.step(loss)


        Args:
            loss: the loss tensor.
            params (optional): the parameters with regard to which we measure
                the loss. These must be provided if the differentiable optimizer
                did not receive a patched model with a view over its own fast
                weights at initialisation. If there is such a model, and params
                are provided, they will overwrite the params of the encapsulated
                model.
            override (optional): a dictionary mapping optimizer settings (i.e.
                those which would be passed to the optimizer constructor or
                provided within parameter groups) to either singleton lists of
                override values, or to a list of override values of length equal
                to the number of parameter groups. If a single override is
                provided for a keyword, it is used for all parameter groups. If
                a list is provided, the ``i``\ th element of the list overrides
                the corresponding setting in the ``i``\ th parameter group. This
                permits the passing of tensors requiring gradient to
                differentiable optimizers for use as optimizer settings. Setting
                override here has highest precedence, i.e. it will override any
                tensors provided as override during the creation of the
                differentiable optimizer, where there is name clash.
            grad_callback: (optional) a single argument function which will be
                applied to a list of gradients of parameters, which respects the
                order specified by ``reference_params``. This can be used to
                apply a function, such as gradient clipping, to all (or a
                subset) of these gradients every time the step function is
                called. This callback overrides the default provided when
                constructing the differentiable optimizer.


        Returns:
            The updated parameters, which will individually have ``grad_fn``\ s
            of their own. If the optimizer has an encapsulated patched model,
            its view over its own fast weights will be updated with these
            params.
        """

        # Deal with override
        if override is not None:
            self._apply_override(override)

        if self._fmodel is None or self._fmodel.fast_params is None:
            if params is None:
                raise ValueError(
                    "params kwarg must be passed to step if the differentiable "
                    "optimizer doesn't have a view on a patched model with "
                    "params."
                )
        else:
            params = self._fmodel.fast_params if params is None else params

        params = list(params)

        grad_targets = [
            p if not p.is_stop_grad() else jit.zeros_like(p).stop_grad() for p in params
        ]

        all_grads = jit.grad(
            loss,
            grad_targets,
            retain_graph=True,  # Jittor does not have allow_unused, retain_graph used here
        )

        if grad_callback is not None:
            all_grads = grad_callback(all_grads)
        elif self._grad_callback is not None:
            all_grads = self._grad_callback(all_grads)

        grouped_grads = []
        for group, mapping in zip(self.param_groups, self._group_to_param_list):
            grads = []
            for i, index in enumerate(mapping):
                group["params"][i] = params[index]
                grads.append(all_grads[index])
            grouped_grads.append(grads)

        self._update(grouped_grads)

        new_params = params[:]
        for group, mapping in zip(self.param_groups, self._group_to_param_list):
            for p, index in zip(group["params"], mapping):
                if self._track_higher_grads:
                    new_params[index] = p
                else:
                    # new_params[index] = p.detach().requires_grad_()
                    new_params[index] = p.detach()
                    new_params[index].start_grad()
        if self._fmodel is not None:
            self._fmodel.update_params(new_params)

        return new_params

    @_abc.abstractmethod
    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        pass


class DifferentiableSGD(DifferentiableOptimizer):
    r"""A differentiable version of the SGD optimizer.

    This optimizer creates a gradient tape as it updates parameters."""

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            momentum = group.get("momentum", 0.0)  # 如果 group 中没有，默认为 0.0
            weight_decay = group.get("weight_decay", 0.0)
            dampening = group.get("dampening", 0.0)
            nesterov = group.get("nesterov", False)
            # 遍历参数和梯度
            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                # 如果 weight_decay 不为 0，则对梯度进行正则化
                if weight_decay != 0:
                    g += weight_decay * p

                # 使用 self.state 管理动量相关状态
                param_state = self.state[group_idx].get(p_idx, {})
                if momentum != 0:
                    # 初始化 momentum_buffer 如果不存在
                    if "momentum_buffer" not in param_state:
                        param_state["momentum_buffer"] = jit.zeros_like(p).stop_grad()

                    buf = param_state["momentum_buffer"]
                    buf *= momentum
                    buf += g * (1 - dampening)

                    if nesterov:
                        # 如果使用 Nesterov 动量
                        g += momentum * buf
                    else:
                        g = buf

                    # 更新状态
                    self.state[group_idx][p_idx] = param_state

                # 最终更新参数
                p -= self.lr * g


class DifferentiableAdam(DifferentiableOptimizer):
    r"""A differentiable version of the Adam optimizer for Jittor.

    This optimizer creates a gradient tape as it updates parameters.
    """

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            amsgrad = group.get("amsgrad", False)
            beta1, beta2 = group.get("betas", (0.9, 0.999))
            weight_decay = group.get("weight_decay", 0.0)
            eps = group.get("eps", 1e-8)

            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                # State initialization
                param_state = self.state[group_idx].get(p_idx, {})
                if not param_state:
                    param_state["step"] = 0
                    param_state["exp_avg"] = jit.zeros_like(p).stop_grad()
                    param_state["exp_avg_sq"] = jit.zeros_like(p).stop_grad()
                    if amsgrad:
                        param_state["max_exp_avg_sq"] = jit.zeros_like(p).stop_grad()

                exp_avg, exp_avg_sq = param_state["exp_avg"], param_state["exp_avg_sq"]
                if amsgrad:
                    max_exp_avg_sq = param_state["max_exp_avg_sq"]

                # Update step count
                param_state["step"] += 1
                step = param_state["step"]

                # Apply weight decay
                if weight_decay != 0:
                    g += weight_decay * p

                # Decay the first and second moment running average coefficient
                exp_avg *= beta1
                exp_avg += (1 - beta1) * g

                exp_avg_sq *= beta2
                exp_avg_sq += (1 - beta2) * (g * g)

                # Bias correction terms
                bias_correction1 = 1 - beta1**step
                bias_correction2 = 1 - beta2**step

                if amsgrad:
                    # Maintain the max of all 2nd moment running avg till now
                    max_exp_avg_sq = jit.maximum(max_exp_avg_sq, exp_avg_sq)
                    param_state["max_exp_avg_sq"] = max_exp_avg_sq
                    denom = (max_exp_avg_sq.sqrt() / jit.sqrt(bias_correction2)) + eps
                else:
                    denom = (exp_avg_sq.sqrt() / jit.sqrt(bias_correction2)) + eps

                step_size = group["lr"] / bias_correction1

                # Update parameters
                p -= step_size * (exp_avg / denom)

                # Save updated state
                self.state[group_idx][p_idx] = param_state


class DifferentiableAdamW(DifferentiableOptimizer):
    r"""A differentiable version of the AdamW optimizer for Jittor.

    This optimizer creates a gradient tape as it updates parameters.
    """

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            amsgrad = group.get("amsgrad", False)
            beta1, beta2 = group.get("betas", (0.9, 0.999))
            weight_decay = group.get("weight_decay", 0.01)  # Typical default for AdamW
            eps = group.get("eps", 1e-8)

            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                # State initialization
                param_state = self.state[group_idx].get(p_idx, {})
                if not param_state:
                    param_state["step"] = 0
                    param_state["exp_avg"] = jit.zeros_like(p).stop_grad()
                    param_state["exp_avg_sq"] = jit.zeros_like(p).stop_grad()
                    if amsgrad:
                        param_state["max_exp_avg_sq"] = jit.zeros_like(p).stop_grad()

                exp_avg, exp_avg_sq = param_state["exp_avg"], param_state["exp_avg_sq"]
                if amsgrad:
                    max_exp_avg_sq = param_state["max_exp_avg_sq"]

                # Apply weight decay directly on the parameter (AdamW specific)
                p *= 1 - group["lr"] * weight_decay

                # Update step count
                param_state["step"] += 1
                step = param_state["step"]

                # Decay the first and second moment running average coefficient
                exp_avg *= beta1
                exp_avg += (1 - beta1) * g

                exp_avg_sq *= beta2
                exp_avg_sq += (1 - beta2) * (g * g)

                # Bias correction terms
                bias_correction1 = 1 - beta1**step
                bias_correction2 = 1 - beta2**step

                if amsgrad:
                    # Maintain the max of all 2nd moment running avg till now
                    max_exp_avg_sq = jit.maximum(max_exp_avg_sq, exp_avg_sq)
                    param_state["max_exp_avg_sq"] = max_exp_avg_sq
                    denom = (max_exp_avg_sq.sqrt() / jit.sqrt(bias_correction2)) + eps
                else:
                    denom = (exp_avg_sq.sqrt() / jit.sqrt(bias_correction2)) + eps

                step_size = group["lr"] / bias_correction1

                # Update parameters
                p -= step_size * (exp_avg / denom)

                # Save updated state
                self.state[group_idx][p_idx] = param_state


class DifferentiableAdadelta(DifferentiableOptimizer):
    r"""A differentiable version of the Adadelta optimizer for Jittor.

    This optimizer creates a gradient tape as it updates parameters.
    """

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            rho = group.get("rho", 0.9)
            eps = group.get("eps", 1e-6)
            weight_decay = group.get("weight_decay", 0.0)

            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                # State initialization
                param_state = self.state[group_idx].get(p_idx, {})
                if not param_state:
                    param_state["step"] = 0
                    param_state["square_avg"] = jit.zeros_like(p).stop_grad()
                    param_state["acc_delta"] = jit.zeros_like(p).stop_grad()

                square_avg = param_state["square_avg"]
                acc_delta = param_state["acc_delta"]

                # Update step count
                param_state["step"] += 1

                # Apply weight decay if needed
                if weight_decay != 0:
                    g += weight_decay * p

                # Update square_avg
                square_avg *= rho
                square_avg += (1 - rho) * (g * g)
                param_state["square_avg"] = square_avg

                # Compute update step
                std = (square_avg + eps).sqrt()
                delta = (acc_delta + eps).sqrt() / std * g

                # Update acc_delta
                acc_delta *= rho
                acc_delta += (1 - rho) * (delta * delta)
                param_state["acc_delta"] = acc_delta

                # Update parameter
                p -= group["lr"] * delta

                # Save updated state
                self.state[group_idx][p_idx] = param_state


class DifferentiableAdagrad(DifferentiableOptimizer):
    r"""A differentiable version of the Adagrad optimizer for Jittor.

    This optimizer creates a gradient tape as it updates parameters.
    """

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            lr = group.get("lr", 1e-2)
            lr_decay = group.get("lr_decay", 0.0)
            weight_decay = group.get("weight_decay", 0.0)
            eps = group.get("eps", 1e-10)

            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                # State initialization
                param_state = self.state[group_idx].get(p_idx, {})
                if not param_state:
                    param_state["step"] = 0
                    param_state["sum"] = jit.zeros_like(p).stop_grad()

                sum_ = param_state["sum"]

                # Update step count
                param_state["step"] += 1
                step = param_state["step"]

                # Apply weight decay if needed
                if weight_decay != 0:
                    g += weight_decay * p

                # Compute adjusted learning rate
                clr = lr / (1 + (step - 1) * lr_decay)

                # Update sum of squared gradients
                sum_ += g * g
                param_state["sum"] = sum_

                # Compute parameter update
                std = sum_.sqrt() + eps
                p -= clr * g / std

                # Save updated state
                self.state[group_idx][p_idx] = param_state


class DifferentiableAdamax(DifferentiableOptimizer):
    r"""A differentiable version of the Adamax optimizer for Jittor.

    This optimizer creates a gradient tape as it updates parameters.
    """

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            lr = group.get("lr", 2e-3)
            betas = group.get("betas", (0.9, 0.999))
            eps = group.get("eps", 1e-8)
            weight_decay = group.get("weight_decay", 0.0)

            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                # State initialization
                param_state = self.state[group_idx].get(p_idx, {})
                if not param_state:
                    param_state["step"] = 0
                    param_state["exp_avg"] = jit.zeros_like(p).stop_grad()
                    param_state["exp_inf"] = jit.zeros_like(p).stop_grad()

                exp_avg = param_state["exp_avg"]
                exp_inf = param_state["exp_inf"]
                beta1, beta2 = betas

                # Update step count
                param_state["step"] += 1
                step = param_state["step"]

                # Apply weight decay
                if weight_decay != 0:
                    g += weight_decay * p

                # Update biased first moment estimate
                exp_avg = exp_avg * beta1 + (1 - beta1) * g
                param_state["exp_avg"] = exp_avg

                # Update the exponentially weighted infinity norm
                exp_inf = jit.maximum(exp_inf * beta2, g.abs())
                param_state["exp_inf"] = exp_inf

                # Bias correction
                bias_correction = 1 - beta1**step
                clr = lr / bias_correction

                # Parameter update
                p -= clr * exp_avg / (exp_inf + eps)

                # Save updated state
                self.state[group_idx][p_idx] = param_state


class DifferentiableASGD(DifferentiableOptimizer):
    r"""A differentiable version of the ASGD optimizer for Jittor.

    This optimizer creates a gradient tape as it updates parameters.
    """

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            lr = group.get("lr", 1e-2)
            lambd = group.get("lambd", 1e-4)
            alpha = group.get("alpha", 0.75)
            t0 = group.get("t0", 1e6)
            weight_decay = group.get("weight_decay", 0.0)

            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                # State initialization
                param_state = self.state[group_idx].get(p_idx, {})
                if not param_state:
                    param_state["step"] = 0
                    param_state["eta"] = lr
                    param_state["mu"] = 1
                    param_state["ax"] = jit.zeros_like(p).stop_grad()

                eta = param_state["eta"]
                mu = param_state["mu"]
                ax = param_state["ax"]

                # Update step count
                param_state["step"] += 1
                step = param_state["step"]

                # Apply weight decay
                if weight_decay != 0:
                    g += weight_decay * p

                # Decay term
                p *= 1 - lambd * eta

                # Update parameter
                p -= eta * g

                # Averaging
                if mu != 1:
                    ax += (p - ax) * mu
                else:
                    ax = p

                # Update eta and mu
                param_state["eta"] = lr / ((1 + lambd * lr * step) ** alpha)
                param_state["mu"] = 1 / max(1, step - t0)

                # Save updated parameter and state
                group["params"][p_idx] = p
                param_state["ax"] = ax
                self.state[group_idx][p_idx] = param_state


class DifferentiableRMSprop(DifferentiableOptimizer):
    r"""A differentiable version of the RMSprop optimizer for Jittor.

    This optimizer creates a gradient tape as it updates parameters.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _warnings.warn(
            "Differentiable RMSprop may suffer from gradient correctness issues. "
            "Consider verifying behavior for specific use cases."
        )

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            lr = group.get("lr", 1e-2)
            alpha = group.get("alpha", 0.99)
            eps = group.get("eps", 1e-8)
            weight_decay = group.get("weight_decay", 0.0)
            momentum = group.get("momentum", 0.0)
            centered = group.get("centered", False)

            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                # State initialization
                param_state = self.state[group_idx].get(p_idx, {})
                if not param_state:
                    param_state["step"] = 0
                    param_state["square_avg"] = jit.zeros_like(p).stop_grad()
                    if momentum > 0:
                        param_state["momentum_buffer"] = jit.zeros_like(p).stop_grad()
                    if centered:
                        param_state["grad_avg"] = jit.zeros_like(p).stop_grad()

                square_avg = param_state["square_avg"]
                param_state["step"] += 1

                # Apply weight decay
                if weight_decay != 0:
                    g += weight_decay * p

                # Update running average of squared gradients
                square_avg = alpha * square_avg + (1 - alpha) * g * g
                param_state["square_avg"] = square_avg

                # Prevent NaNs for zero values
                if jit.any(square_avg == 0):
                    square_avg += eps

                # Calculate centered RMSprop if applicable
                if centered:
                    grad_avg = param_state["grad_avg"]
                    grad_avg = alpha * grad_avg + (1 - alpha) * g
                    param_state["grad_avg"] = grad_avg
                    avg = (square_avg - grad_avg * grad_avg).sqrt() + eps
                else:
                    avg = square_avg.sqrt() + eps

                # Momentum updates if enabled
                if momentum > 0:
                    buf = param_state.get(
                        "momentum_buffer", jit.zeros_like(p).stop_grad()
                    )
                    buf = momentum * buf + g / avg
                    param_state["momentum_buffer"] = buf
                    p -= lr * buf
                else:
                    p -= lr * g / avg

                # Save updated parameter and state
                group["params"][p_idx] = p
                self.state[group_idx][p_idx] = param_state


class DifferentiableRprop(DifferentiableOptimizer):
    r"""A differentiable version of the Rprop optimizer for Jittor.

    This optimizer creates a gradient tape as it updates parameters.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _warnings.warn(
            "Differentiable Rprop correctly yields zero second-order gradients, "
            "as only the sign of the gradient is used in updates. Future versions "
            "may include higher-order gradient approximations."
        )

    def _update(self, grouped_grads: _GroupedGradsType, **kwargs) -> None:
        zipped = zip(self.param_groups, grouped_grads)

        for group_idx, (group, grads) in enumerate(zipped):
            etaminus, etaplus = group.get("etas", (0.5, 1.2))
            step_size_min, step_size_max = group.get("step_sizes", (1e-6, 50))

            for p_idx, (p, g) in enumerate(zip(group["params"], grads)):
                if g is None or p.is_stop_grad():
                    continue

                if g.is_sparse:
                    raise RuntimeError("Rprop does not support sparse gradients")

                # State initialization
                param_state = self.state[group_idx].get(p_idx, {})
                if not param_state:
                    param_state["step"] = 0
                    param_state["prev"] = jit.zeros_like(p).stop_grad()
                    param_state["step_size"] = jit.full_like(
                        p, group.get("lr", 1e-2)
                    ).stop_grad()

                prev_grad = param_state["prev"]
                step_size = param_state["step_size"]

                param_state["step"] += 1

                # Determine gradient sign
                sign = (g * prev_grad).sign()
                sign_positive = sign > 0
                sign_negative = sign < 0

                # Update step sizes
                step_size = jit.where(sign_positive, step_size * etaplus, step_size)
                step_size = jit.where(sign_negative, step_size * etaminus, step_size)
                step_size = step_size.clamp(step_size_min, step_size_max)
                param_state["step_size"] = step_size

                # Zero out gradient where sign is negative
                g = jit.where(sign_negative, jit.zeros_like(g), g)

                # Update parameters
                p -= g.sign() * step_size

                # Save state
                param_state["prev"] = g.clone().stop_grad()
                self.state[group_idx][p_idx] = param_state


_OptMappingType = _typing.Dict[
    jit.optim.Optimizer, _typing.Type[DifferentiableOptimizer]
]
# _opt_mapping: _OptMappingType = {
#     jit.optim.Adadelta: DifferentiableAdadelta,
#     jit.optim.Adagrad: DifferentiableAdagrad,
#     jit.optim.Adam: DifferentiableAdam,
#     jit.optim.AdamW: DifferentiableAdamW,
#     jit.optim.Adamax: DifferentiableAdamax,
#     jit.optim.ASGD: DifferentiableASGD,
#     jit.optim.RMSprop: DifferentiableRMSprop,
#     jit.optim.Rprop: DifferentiableRprop,
#     jit.optim.SGD: DifferentiableSGD,
# }

# 获取实际存在的优化器
available_optimizers = {attr for attr in dir(jit.optim) if not attr.startswith("_")}

# 定义优化器映射
_opt_mapping: _OptMappingType = {
    "Adadelta": DifferentiableAdadelta,
    "Adagrad": DifferentiableAdagrad,
    "Adam": DifferentiableAdam,
    "AdamW": DifferentiableAdamW,
    "Adamax": DifferentiableAdamax,
    "ASGD": DifferentiableASGD,
    "RMSprop": DifferentiableRMSprop,
    "Rprop": DifferentiableRprop,
    "SGD": DifferentiableSGD,
}

# 移除不存在的优化器
_opt_mapping = {
    getattr(jit.optim, k): v
    for k, v in _opt_mapping.items()
    if k in available_optimizers
}

# print("Updated optimizer mapping:", _opt_mapping)


def get_diff_optim(
    opt: jit.optim.Optimizer,
    reference_params: _typing.Iterable[jit.Var],
    fmodel: _typing.Optional[_patch._MonkeyPatchBase] = None,
    device: _typing.Optional[str] = None,
    override: _typing.Optional[_OverrideType] = None,
    track_higher_grads: bool = True,
    **kwargs,
) -> DifferentiableOptimizer:
    r"""Construct/initialize a differentiable version of an existing optimizer.

    Args:
        opt: an existing optimizer, assumed to be an instance of
            ``jittor.optim.Optimizer``, of a supported type which is either defined
            in ``jittor.optim``, or a custom implementation which has been added to
            higher at runtime by using ``higher.register_optim``. We assume this
            optimizer tracks the parameters (or some subset thereof) of a single
            ``jittor.Module`` instance, with support for parameter groups.
        reference_params: the parameters of the module tracked by ``opt``, as
            returned by ``module.parameters()``.
        fmodel (optional): a patched version of the ``module`` tracked by ``opt``.
            It is assumed this patched instance has a view on its latest fast
            weights through ``fmodel.parameters()``. If provided, it is not
            necessary to pass the fast weights explicitly to the differentiable
            optimizer's ``step`` function via the keyword arg ``params``. If not
            provided, the fast weights to update must be provided to ``step``.
        device (optional): the device to cast the optimizer state to when
            creating the differentiable optimizer. If not provided, the same
            device as used for the parameters tracked by ``opt`` will be used.
        override (optional): a dictionary mapping optimizer settings (i.e.
            those which would be passed to the optimizer constructor or
            provided within parameter groups) to either singleton lists of
            override values, or to a list of override values of length equal to
            the number of parameter groups. If a single override is provided for
            a keyword, it is used for all parameter groups. If a list is
            provided, the ``i``\ th element of the list overrides the corresponding
            setting in the ``i``\ th parameter group. This permits the passing of
            tensors requiring gradient to differentiable optimizers for use as
            optimizer settings.
        track_higher_grads: if True, during unrolled optimization the graph be
            retained, and the fast weights will bear grad funcs, so as to permit
            backpropagation through the optimization process. Setting this to
            False allows the returned differentiable optimizer to be used in
            "test mode", without potentially tracking higher order gradients.
            This can be useful when running the training loop at test time,
            e.g. in k-shot learning experiments, without incurring a significant
            memory overhead.

    Returns:
        An initialized ``DifferentiableOptimizer`` instance of the right subtype.
    """
    if type(opt) in _opt_mapping:
        return _opt_mapping[type(opt)](
            opt,
            reference_params,
            fmodel=fmodel,
            device=device,
            override=override,
            track_higher_grads=track_higher_grads,
            **kwargs,
        )
    else:
        raise ValueError(
            "Optimizer type {} not supported by higher yet.".format(type(opt))
        )


def create_diff_optim(
    opt_type: _typing.Type[jit.optim.Optimizer],
    opt_kwargs: _typing.Optional[_typing.Dict[str, _typing.Any]] = None,
    params: _typing.Optional[_typing.List[jit.Var]] = None,
    fmodel: _typing.Optional[_patch._MonkeyPatchBase] = None,
    device: _typing.Optional[str] = None,
    override: _typing.Optional[_OverrideType] = None,
    track_higher_grads: bool = True,
    **kwargs,
) -> DifferentiableOptimizer:
    r"""Construct a differentiable version of an new optimizer.

    Args:
        opt_type: the type (constructor) for a jittor.optim.Optimizer subtype
            from amongst the types supported by the library, or registered with
            it a runtime.
        opt_kwargs: a dictionary of keywords to be passed to the optimizer
            constructor.
        params (optional): a list of (fast) weights which the differentiable
            optimizer will update. These must be provided if fmodel is not
            provided. If both, these will be used in lieu. These will only
            be used for shape inference when initializing the optimizer.
            This argument can also take the same format as parameter groups,
            i.e. an iterable over dictionaries which contain the 'params' key
            with fast weights as value, and group-specific hyperparameters.
        fmodel (optional): a patched version of the ``module`` tracked by ``opt``.
            It is assumed this patched instance has a view on its latest fast
            weights through ``fmodel.parameters()``. If provided, it is not
            necessary to pass the fast weights explicitly to the differentiable
            optimizer's ``step`` function via the keyword arg ``params``. If not
            provided, the fast weights to update must be provided to ``step``.
        device (optional): the device to cast the optimizer state to when
            creating the differentiable optimizer. If not provided, the same
            device as used for the parameters tracked by ``opt`` will be used.
        override (optional): a dictionary mapping optimizer settings (i.e.
            those which would be passed to the optimizer constructor or
            provided within parameter groups) to either singleton lists of
            override values, or to a list of override values of length equal to
            the number of parameter groups. If a single override is provided for
            a keyword, it is used for all parameter groups. If a list is
            provided, the ``i``\ th element of the list overrides the corresponding
            setting in the ``i``\ th parameter group. This permits the passing of
            tensors requiring gradient to differentiable optimizers for use as
            optimizer settings.
        track_higher_grads: if True, during unrolled optimization the graph be
            retained, and the fast weights will bear grad funcs, so as to permit
            backpropagation through the optimization process. Setting this to
            False allows the returned differentiable optimizer to be used in
            "test mode", without potentially tracking higher order gradients.
            This can be useful when running the training loop at test time,
            e.g. in k-shot learning experiments, without incurring a significant
            memory overhead.

    Returns:
        An initialized ``DifferentiableOptimizer`` instance of the right subtype.
    """

    if opt_type in _opt_mapping:
        if params is not None:
            params = list(params)
            if isinstance(params[0], dict):
                dummy = [
                    {
                        k: jit.zeros_like(v, requires_grad=True) if k == "params" else v
                        for k, v in group.items()
                    }
                    for group in params
                ]
            else:
                dummy = [jit.zeros_like(p, requires_grad=True) for p in params]
        elif fmodel is not None:
            dummy = [jit.zeros_like(p, requires_grad=True) for p in fmodel.parameters()]
        else:
            raise ValueError("Must specify one of fmodel or params in kwargs.")

        opt_kwargs = {} if opt_kwargs is None else opt_kwargs
        opt = opt_type(dummy, **opt_kwargs)

        return _opt_mapping[opt_type](
            opt,
            dummy,
            fmodel=fmodel,
            device=device,
            override=override,
            track_higher_grads=track_higher_grads,
            **kwargs,
        )
    else:
        raise ValueError(
            "Optimizer type {} not supported by higher yet.".format(opt_type)
        )


def register_optim(
    optim_type: jit.optim.Optimizer,
    diff_optim_type: _typing.Type[DifferentiableOptimizer],
) -> None:
    r"""Registers a new optimizer type for use with higher functions.

    Args:
        optim_type: the type of a new optimizer, assumed to be an instance of
            ``jittor.optim.Optimizer``.
        diff_optim_type: the type of a new differentiable optimizer, assumed to
            be an instance of ``higher.optim.DifferentiableOptimizer`` with
            functionally equivalent logic to ``optim_type``.
    """
    _opt_mapping[optim_type] = diff_optim_type


def get_trainable_opt_params(
    opt: jit.optim.Optimizer, device: _typing.Optional[str] = None
) -> _OverrideType:
    r"""Get an override dictionary from an optimizer instance.

    Args:
        opt: the optimizer to obtain an override dictionary from.
        device (optional): the device to cast the learnable tensors to.

    Returns:
        A dictionary of the format expected for the override kwarg of
        differentiable optimizers. It is initialized with trainable tensors
        with as values those float and int hyperparameters found in the
        optimizer's parameter groups (or stuctures containing these).
        Heuristically, hyperparameters containing mixtures of differentiable
        and non-differentiable types will be ignored (and must be manually
        specified when constructing an override dict).
    """
    override: _OverrideType = _collections.defaultdict(list)

    def map_fn(x: _typing.Union[jit.Var, int, float]) -> jit.Var:
        if isinstance(x, jit.Var):
            return x.clone().detach()
        else:
            return jit.array(float(x)).stop_grad(False)

    for group in opt.param_groups:
        for k, v in group.items():
            if k == "params":
                # Ignore actual model parameters tracked by optim
                continue

            # Ignore hyperparameters that aren't structures containing ints
            # or floats
            if all(
                isinstance(x, int) or isinstance(x, float) for x in _utils.flatten(v)
            ):
                override[k].append(_utils._recursive_map(v, map_fn))

    return override


def apply_trainable_opt_params(
    opt: jit.optim.Optimizer, override: _OverrideType
) -> None:
    r"""Apply learned hyperparameters back to original optimizer.

    Args:
        opt: the original optimizer. The hyperparameters in its parameter groups
            will be modified in place.
        override: dictionary of the format used for the override kwarg of
            differentiable optimizers.
    """
    for k, v in override.items():
        # Sanity check
        if (len(v) != 1) and (len(v) != len(opt.param_groups)):
            raise ValueError(
                "Mismatch between the number of override tensors for "
                "optimizer parameter {} and the number of "
                "parameter groups.".format(k)
            )
        for group_idx, group in enumerate(opt.param_groups):
            replacement = v[0] if len(v) == 1 else v[group_idx]
            group[k] = _recursive_apply(replacement, group[k])


## Local utility functions
# TODO(egrefen): use funcs below instead of x._add, in diffopt
def _add(
    tensor: jit.Var,
    a1: _typing.Union[float, int, jit.Var],
    a2: _typing.Optional[jit.Var] = None,
) -> jit.Var:
    if a2 is None:
        value: _typing.Union[jit.Var, float] = 1.0
        other = a1
    else:
        value = a1
        other = a2
    return tensor + (value * other)


def _addcdiv(
    tensor: jit.Var,
    a1: _typing.Union[float, int, jit.Var],
    a2: jit.Var,
    a3: _typing.Optional[jit.Var] = None,
) -> jit.Var:
    if a3 is None:
        value: _typing.Union[jit.Var, float] = 1.0
        tensor1 = a1
        tensor2 = a2
    else:
        value = a1
        tensor1 = a2
        tensor2 = a3
    return tensor + value * (tensor1 / tensor2)


def _addcmul(
    tensor: jit.Var,
    a1: _typing.Union[float, int, jit.Var],
    a2: jit.Var,
    a3: _typing.Optional[jit.Var] = None,
) -> jit.Var:
    if a3 is None:
        value: _typing.Union[jit.Var, float] = 1.0
        tensor1 = a1
        tensor2 = a2
    else:
        value = a1
        tensor1 = a2
        tensor2 = a3
    return tensor + (value * tensor1 * tensor2)


# TODO(egrefen): this probably could be refactored into utils
def _recursive_apply(
    replacement: _typing.Union[list, tuple, dict, set, jit.Var],
    target: _typing.Union[jit.Var, int, float],
) -> _typing.Union[jit.Var, int, float]:
    if not isinstance(replacement, type(target)):
        if isinstance(replacement, jit.Var) and not _utils._is_container(target):
            return type(target)(replacement.item())
        raise ValueError(
            "Expected an non-container type for target, but got {} with value "
            "{}".format(type(target), target)
        )
    elif isinstance(replacement, jit.Var) and isinstance(target, jit.Var):
        replacement = replacement.to(target.device)
        target.data = replacement.data
        return target
    if isinstance(target, list):
        return type(target)(
            [_recursive_apply(r, t) for r, t in zip(replacement, target)]
        )
    elif isinstance(target, tuple):
        return type(target)(
            [_recursive_apply(r, t) for r, t in zip(replacement, target)]
        )
    elif isinstance(replacement, dict) and isinstance(target, dict):
        return type(target)(
            {
                k: _recursive_apply(r, t)
                for (_, r), (k, t) in zip(replacement.items(), target.items())
            }
        )
    elif isinstance(target, set):
        return type(target)(
            {_recursive_apply(r, t) for r, t in zip(replacement, target)}
        )
    else:
        raise ValueError(
            "Couldn't apply replacement of type {} to target of type "
            "{}".format(type(replacement), type(target))
        )



import jittor as jit
from typing import List, Callable, Dict


class ResultStore:
    """
    A simple class to store and manage intermediate results of hyper-gradient computation.
    """

    def __init__(self):
        self.results = []

    def add(self, name: str, result: Dict):
        """
        Add a result to the store.

        Parameters
        ----------
        name : str
            The name of the result (e.g., 'gradient_operator_results_0').
        result : Dict
            The result dictionary to store.
        """

        self.results.append({name: result})

    def clear(self):
        """Clear all stored results."""
        self.results = []

    def get_results(self) -> List[Dict]:
        """Retrieve all stored results."""
        return self.results


class DynamicalSystemRules:
    """
    A class to store and manage gradient operator rules.
    """

    # Default static gradient operator order
    _gradient_order = [
        ["GDA", "DI"],
        ["DM", "NGD"],
    ]

    @staticmethod
    def get_gradient_order() -> List[List[str]]:
        """
        Get the current gradient operator order.

        Returns
        -------
        List[List[str]]
            The current gradient operator order.
        """
        return DynamicalSystemRules._gradient_order

    @staticmethod
    def set_gradient_order(new_order: List[List[str]]):
        """
        Set a new gradient operator order.

        Parameters
        ----------
        new_order : List[List[str]]
            The new gradient operator order to set.

        Raises
        ------
        ValueError
            If the new order is invalid.
        """
        if not isinstance(new_order, list) or not all(
            isinstance(group, list) for group in new_order
        ):
            raise ValueError("Gradient order must be a list of lists.")
        DynamicalSystemRules._gradient_order = new_order


class HyperGradientRules:
    """
    A class to store and manage gradient operator rules.
    """

    # Default static gradient operator order
    _gradient_order = [
        ["PTT", "FOA", "RGT"],
        ["IAD", "RAD", "FD", "IGA"],
        ["CG", "NS"],
    ]

    @staticmethod
    def get_gradient_order() -> List[List[str]]:
        """
        Get the current gradient operator order.

        Returns
        -------
        List[List[str]]
            The current gradient operator order.
        """
        return HyperGradientRules._gradient_order

    @staticmethod
    def set_gradient_order(new_order: List[List[str]]):
        """
        Set a new gradient operator order.

        Parameters
        ----------
        new_order : List[List[str]]
            The new gradient operator order to set.

        Raises
        ------
        ValueError
            If the new order is invalid.
        """
        if not isinstance(new_order, list) or not all(
            isinstance(group, list) for group in new_order
        ):
            raise ValueError("Gradient order must be a list of lists.")
        HyperGradientRules._gradient_order = new_order


def l2_reg(parameters):
    """
    Compute the L2 regularization loss for a list of parameters.

    Parameters
    ----------
    parameters : Iterable[jittor.Var]
        Model parameters for which the L2 regularization is computed.

    Returns
    -------
    jittor.Var
        L2 regularization loss.
    """
    loss = 0
    for w in parameters:
        loss += (w**2).sum()
    return loss


def grad_unused_zero(output, inputs, retain_graph=False):
    """
    Compute gradients for inputs with respect to the output, filling missing gradients with zeros.

    Parameters
    ----------
    output : jittor.Var
        The output tensor to compute gradients for.
    inputs : List[jittor.Var]
        The input tensors to compute gradients with respect to.
    retain_graph : bool, optional
        Whether to retain the computation graph, by default False.

    Returns
    -------
    Tuple[jittor.Var]
        Gradients with respect to the inputs, with zeros for unused gradients.
    """
    grads = jit.grad(output, inputs, retain_graph=retain_graph)

    def grad_or_zeros(grad, var):
        return jit.zeros_like(var) if grad is None or (jit.isnan(grad).any()) else grad

    return tuple(grad_or_zeros(g, v) for g, v in zip(grads, inputs))


def list_tensor_matmul(list1, list2):
    """
    Compute the element-wise multiplication and sum of two lists of tensors.

    Parameters
    ----------
    list1 : List[jittor.Var]
        The first list of tensors.
    list2 : List[jittor.Var]
        The second list of tensors.

    Returns
    -------
    jittor.Var
        The resulting scalar from element-wise multiplication and summation.
    """
    out = 0
    for t1, t2 in zip(list1, list2):
        out += (t1 * t2).sum()
    return out


def list_tensor_norm(list_tensor, p=2):
    """
    Compute the p-norm of a list of tensors.

    Parameters
    ----------
    list_tensor : List[jittor.Var]
        The list of tensors to compute the norm for.
    p : float, optional
        The order of the norm, by default 2 (Euclidean norm).

    Returns
    -------
    jittor.Var
        The computed p-norm of the list of tensors.

    Raises
    ------
    ValueError
        If the list of tensors is empty.
    """
    norm = 0
    for t in list_tensor:
        norm += (t.abs() ** p).sum()
    return norm ** (1 / p)


def require_model_grad(model=None):
    """
    Ensure all model parameters require gradients.

    Parameters
    ----------
    model : jittor.Module
        The model to check and update parameters.

    Raises
    ------
    AssertionError
        If the model is not defined.
    """
    assert model is not None, "The module is not defined!"
    for param in model.parameters():
        if param.is_stop_grad():
            param = param.clone()


def update_grads(grads, model):
    """
    Update the custom_grad attribute of the model's parameters.

    Parameters
    ----------
    grads : List[jittor.Var]
        Gradients to be applied to the parameters.
    model : jittor.Module
        Model whose parameters will be updated.
    """
    for p, x in zip(grads, model.parameters()):
        if not hasattr(x, "_custom_grad"):
            x._custom_grad = p.clone()
        else:
            x._custom_grad += p



def manual_update(optimizer, variables):
    """
    Manually update variables using gradients stored in _custom_grad.

    Parameters
    ----------
    optimizer : jittor.optim.Optimizer
        The Jittor optimizer instance.
    variables : List[jittor.Var]
        A list of Jittor variables to be updated.

    Raises
    ------
    AttributeError
        If a variable does not have the '_custom_grad' attribute.
    """
    variable_ids = {id(var) for var in variables} 
    for group in optimizer.param_groups:
        lr = group.get("lr", optimizer.lr)

        for param in group["params"]:
            if id(param) in variable_ids: 
                if not hasattr(param, "_custom_grad"):
                    raise AttributeError(
                        f"Variable '{param.name}' does not have '_custom_grad'. "
                        f"Ensure gradients are precomputed and stored before updating."
                    )

                grad = param._custom_grad
                if grad.shape != param.shape:
                    raise ValueError(
                        f"Gradient shape {grad.shape} does not match parameter shape {param.shape} "
                        f"for variable '{param.name}'"
                    )

                param -= lr * grad

                param._custom_grad *= 0


# def manual_update(optimizer, variables):
#     """
#     Manually update variables using gradients stored in _custom_grad.

#     Parameters
#     ----------
#     optimizer : jittor.optim.Optimizer
#         The Jittor optimizer instance.
#     variables : List[jittor.Var]
#         A list of Jittor variables to be updated.

#     Raises
#     ------
#     AttributeError
#         If a variable does not have the '_custom_grad' attribute.
#     """
#     print(len(variables))
#     print(len(optimizer.param_groups))
#     for group in optimizer.param_groups:
#         lr = group.get("lr", optimizer.lr)

#         for param in group["params"]:
#             # print(variables)
#             print(param.shape)
#             print(param._custom_grad.shape)
#             if param in variables:
#                 if not hasattr(param, "_custom_grad"):
#                     raise AttributeError(
#                         f"Variable '{param.name}' does not have '_custom_grad'. "
#                         f"Ensure gradients are precomputed and stored before updating."
#                     )

#                 grad = param._custom_grad
#                 param -= lr * grad
#                 param._custom_grad *= 0




def update_tensor_grads(hparams, grads):
    """
    Update gradients for Jittor variables manually.

    Parameters
    ----------
    hparams : List[jittor.Var]
        List of Jittor variables representing the hyperparameters.
    grads : List[jittor.Var]
        List of gradients corresponding to the hyperparameters.

    Raises
    ------
    ValueError
        If a variable is stop_grad and cannot be updated.
    """
    for l, g in zip(hparams, grads):
        if l.is_stop_grad():
            raise ValueError(f"Variable {l.name()} is stop_grad and cannot be updated.")
        if not hasattr(l, "_custom_grad"):
            l._custom_grad = g.clone().detach()
        else:
            l._custom_grad += g


def stop_grads(grads):
    """
    Detach and stop gradient computation for a list of gradients.

    Parameters
    ----------
    grads : List[jittor.Var]
        The gradients to process.

    Returns
    -------
    List[jittor.Var]
        Detached gradients with requires_grad set to False.
    """
    return [(grad.detach().stop_grad() if grad is not None else grad) for grad in grads]


def average_grad(model, batch_size):
    """
    Divide the gradients of all model parameters by the batch size.

    Parameters
    ----------
    model : jittor.Module
        The model whose gradients need to be averaged.
    batch_size : int
        The batch size to divide gradients by.
    """
    for param in model.parameters():
        if param.opt_grad() is not None:
            param.opt_grad().update(param.opt_grad() / batch_size)


def stop_model_grad(model=None):
    """
    Stop gradient computation for all parameters in a model.

    Parameters
    ----------
    model : jittor.Module
        The model to stop gradients for.

    Raises
    ------
    AssertionError
        If the model is not defined.
    """
    assert model is not None, "The module is not defined!"
    for param in model.parameters():
        param.stop_grad()


def cat_list_to_tensor(list_tx):
    """
    Concatenate a list of tensors into a single flattened tensor.

    Parameters
    ----------
    list_tx : List[jittor.Var]
        The list of tensors to concatenate.

    Returns
    -------
    jittor.Var
        A single flattened tensor.
    """
    return jit.concat([xx.flatten() for xx in list_tx])


def copy_parameter_from_list(y, z):
    """
    Copy parameters from a list to the parameters of a Jittor model.

    Parameters
    ----------
    y : jittor.Module
        Jittor model with parameters to be updated.
    z : List[jittor.Var]
        List of variables to copy from.

    Returns
    -------
    jittor.Module
        Updated model.
    """
    for p, q in zip(y.parameters(), z):
        p.update(q.clone().detach())
        p.requires_grad = True
    return y


def get_outer_gradients(outer_loss, params, hparams, retain_graph=True):
    """
    Compute gradients of the outer-level loss with respect to parameters and hyperparameters.

    Parameters
    ----------
    outer_loss : jittor.Var
        The scalar loss from the outer-level optimization problem.
    params : List[jittor.Var]
        The list of parameters for which gradients with respect to the outer loss are computed.
    hparams : List[jittor.Var]
        The list of hyperparameters for which gradients with respect to the outer loss are computed.
    retain_graph : bool, optional
        Whether to retain the computation graph after computing the gradients, by default True.

    Returns
    -------
    Tuple[List[jittor.Var], List[jittor.Var]]
        Gradients with respect to parameters and hyperparameters.
    """
    grad_outer_w = grad_unused_zero(outer_loss, params, retain_graph=retain_graph)
    grad_outer_hparams = grad_unused_zero(
        outer_loss, hparams, retain_graph=retain_graph
    )
    return grad_outer_w, grad_outer_hparams


def custom_grad(outputs, inputs, grad_outputs=None, retain_graph=False):
    """
    Compute the vector-Jacobian product for Jittor, mimicking PyTorch's autograd.grad.

    Parameters
    ----------
    outputs : Sequence[jittor.Var]
        Outputs of the differentiated function.
    inputs : Sequence[jittor.Var]
        Inputs with respect to which the gradient will be computed.
    grad_outputs : Sequence[jittor.Var], optional
        Gradients with respect to the outputs, by default None.
    retain_graph : bool, optional
        Whether to retain the computation graph after computing the gradients, by default False.

    Returns
    -------
    List[jittor.Var]
        Gradients with respect to the inputs.
    """
    if not isinstance(outputs, (tuple, list)):
        outputs = (outputs,)
    if grad_outputs is None:
        grad_outputs = [jit.ones_like(output) for output in outputs]
    elif not isinstance(grad_outputs, (tuple, list)):
        grad_outputs = (grad_outputs,)

    assert len(outputs) == len(
        grad_outputs
    ), "outputs and grad_outputs must have the same length."

    total_output = sum(
        (output * grad_output).sum()
        for output, grad_output in zip(outputs, grad_outputs)
    )
    grads = jit.grad(total_output, inputs, retain_graph=retain_graph)
    return grads


def neumann(params, hparams, upper_loss, lower_loss, k, fp_map, tol=1e-10):
    """
    Compute hyperparameter gradients using the Neumann series approximation.

    Parameters
    ----------
    params : List[jittor.Var]
        List of parameters for the lower-level optimization problem.
    hparams : List[jittor.Var]
        List of hyperparameters for the upper-level optimization problem.
    upper_loss : jittor.Var
        Loss function for the upper-level problem.
    lower_loss : jittor.Var
        Loss function for the lower-level problem.
    k : int
        Number of iterations for the Neumann series approximation.
    fp_map : Callable
        Fixed-point map function that computes updates to lower-level parameters.
    tol : float, optional
        Tolerance for early stopping based on convergence, by default 1e-10.

    Returns
    -------
    List[jittor.Var]
        Hyperparameter gradients computed using the Neumann series approximation.
    """
    grad_outer_w, grad_outer_hparams = get_outer_gradients(upper_loss, params, hparams)
    w_mapped = fp_map(params, lower_loss)
    vs, gs = grad_outer_w, grad_outer_w
    gs_vec = cat_list_to_tensor(gs)
    for i in range(k):
        gs_prev_vec = gs_vec
        vs = custom_grad(w_mapped, params, grad_outputs=vs, retain_graph=True)
        gs = [g + v for g, v in zip(gs, vs)]
        gs_vec = cat_list_to_tensor(gs)
        if float(jit.norm(gs_vec - gs_prev_vec)) < tol:
            break

    grads = custom_grad(w_mapped, hparams, grad_outputs=gs)
    grads = [g + v for g, v in zip(grads, grad_outer_hparams)]
    return grads


def conjugate_gradient(
    params, hparams, upper_loss, lower_loss, K, fp_map, tol=1e-10, stochastic=False
):
    """
    Compute hyperparameter gradients using the Conjugate Gradient method.

    Parameters
    ----------
    params : List[jittor.Var]
        List of parameters for the lower-level optimization problem.
    hparams : List[jittor.Var]
        List of hyperparameters for the upper-level optimization problem.
    upper_loss : jittor.Var
        Loss function for the upper-level problem.
    lower_loss : jittor.Var
        Loss function for the lower-level problem.
    K : int
        Maximum number of iterations for the Conjugate Gradient method.
    fp_map : Callable
        Fixed-point map function that computes updates to lower-level parameters.
    tol : float, optional
        Tolerance for early stopping based on convergence, by default 1e-10.
    stochastic : bool, optional
        If True, recompute the fixed-point map during each iteration, by default False.

    Returns
    -------
    List[jittor.Var]
        Hyperparameter gradients computed using the Conjugate Gradient method.
    """
    grad_outer_w, grad_outer_hparams = get_outer_gradients(upper_loss, params, hparams)
    if not stochastic:
        w_mapped = fp_map(params, lower_loss)

    def dfp_map_dw(xs):
        if stochastic:
            w_mapped_in = fp_map(params, lower_loss)
            Jfp_mapTv = custom_grad(
                w_mapped_in, params, grad_outputs=xs, retain_graph=False
            )
        else:
            Jfp_mapTv = custom_grad(
                w_mapped, params, grad_outputs=xs, retain_graph=True
            )
        return [v - j for v, j in zip(xs, Jfp_mapTv)]

    vs = cg_step(dfp_map_dw, grad_outer_w, max_iter=K, epsilon=tol)
    if stochastic:
        w_mapped = fp_map(params, lower_loss)

    grads = custom_grad(w_mapped, hparams, grad_outputs=vs)
    grads = [g + v for g, v in zip(grads, grad_outer_hparams)]
    return grads


def cg_step(Ax, b, max_iter=100, epsilon=1.0e-5):
    """
    Perform Conjugate Gradient (CG) optimization to solve Ax = b.

    Parameters
    ----------
    Ax : Callable
        Function that computes the matrix-vector product Ax for a given x.
    b : List[jittor.Var]
        Right-hand side of the equation Ax = b.
    max_iter : int, optional
        Maximum number of iterations for the CG method, by default 100.
    epsilon : float, optional
        Convergence threshold for the residual norm, by default 1e-5.

    Returns
    -------
    List[jittor.Var]
        Solution vector x that approximately solves Ax = b.
    """
    x_last = [jit.zeros_like(bb) for bb in b]
    r_last = [jit.zeros_like(bb) + bb for bb in b]
    p_last = [jit.zeros_like(rr) + rr for rr in r_last]

    for _ in range(max_iter):
        Ap = Ax(p_last)
        Ap_vec = cat_list_to_tensor(Ap)
        p_last_vec = cat_list_to_tensor(p_last)
        r_last_vec = cat_list_to_tensor(r_last)
        rTr = jit.sum(r_last_vec * r_last_vec)
        pAp = jit.sum(p_last_vec * Ap_vec)
        alpha = rTr / pAp

        x = [xx + alpha * pp for xx, pp in zip(x_last, p_last)]
        r = [rr - alpha * pp for rr, pp in zip(r_last, Ap)]
        r_vec = cat_list_to_tensor(r)

        if float(jit.norm(r_vec)) < epsilon:
            break

        beta = jit.sum(r_vec * r_vec) / rTr
        p = [rr + beta * pp for rr, pp in zip(r, p_last)]

        x_last = x
        p_last = p
        r_last = r

    return x_last



from jittor import Module
from typing import List, Callable, Dict
from ..higher_jit.patch import _MonkeyPatchBase


from boat_jit.operation_registry import register_class
from boat_jit.hyper_ol.hyper_gradient import HyperGradient


@register_class
class PTT(HyperGradient):
    """
    Computes the hyper-gradient of the upper-level variables using Pessimistic Trajectory Truncation (PTT) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        List of variables optimized with the lower-level objective.
    ul_var : List[jittor.Var]
        List of variables optimized with the upper-level objective.
    solver_config : Dict[str, Any]
        Dictionary containing solver configurations, including:
        - "hyper_op" (List[str]): Indicates if PTT is used in the hyper-gradient operations.

    References
    ----------
    [1] Liu R., Liu Y., Zeng S., et al. "Towards gradient-based bilevel optimization with non-convex followers and beyond," in NeurIPS, 2021.
    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict,
    ):
        super(PTT, self).__init__(
            ll_objective,
            ul_objective,
            ul_model,
            ll_model,
            ll_var,
            ul_var,
            solver_config,
        )
        self.truncate_max_loss_iter = "PTT" in solver_config["hyper_op"]

    def compute_gradients(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        max_loss_iter: int = 0,
        hyper_gradient_finished: bool = False,
        next_operation: str = None,
        **kwargs
    ):
        """
        Compute the hyper-gradients of the upper-level variables with the data from feed_dict and patched models.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            It typically includes training data, targets, and other information required to compute the LL objective.

        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            It typically includes validation data, targets, and other information required to compute the UL objective.

        auxiliary_model : _MonkeyPatchBase
            A patched lower model wrapped by the `higher` library.
            It serves as the lower-level model for optimization.

        max_loss_iter : int, optional
            The number of iterations used for backpropagation, by default 0.

        next_operation : str, optional
            The next operator for the calculation of the hypergradient, by default None.

        hyper_gradient_finished : bool, optional
            A boolean flag indicating whether the hypergradient computation is finished, by default False.

        Returns
        -------
        Dict
            A dictionary containing updated feed_dict, auxiliary model, and gradient computation results.
        """
        assert (
            hyper_gradient_finished is False
        ), "Hypergradient computation should not be finished"
        assert self.truncate_max_loss_iter and (
            max_loss_iter > 0
        ), "With PTT operation, 'max_loss_iter' should be greater than 0"
        assert next_operation is not None, "Next operation should be defined"
        lower_model_params = kwargs.get(
            "lower_model_params", list(auxiliary_model.parameters(time=max_loss_iter))
        )
        return {
            "ll_feed_dict": ll_feed_dict,
            "ul_feed_dict": ul_feed_dict,
            "auxiliary_model": auxiliary_model,
            "max_loss_iter": max_loss_iter,
            "hyper_gradient_finished": False,
            "lower_model_params": lower_model_params,
            **kwargs,
        }



from jittor import Module
from ..higher_jit.patch import _MonkeyPatchBase
from ..higher_jit.optim import DifferentiableOptimizer
from typing import Dict, Any, Callable

from boat_jit.operation_registry import register_class
from boat_jit.dynamic_ol.dynamical_system import DynamicalSystem


@register_class
class DI(DynamicalSystem):
    """
    Implements the lower-level optimization procedure for Dynamic Initialization [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    lower_loop : int
        The number of iterations for the lower-level optimization process.
    solver_config : Dict[str, Any]
        A dictionary containing configurations for the optimization solver, including
        hyperparameters and specific settings for NGD and GDA.

    References
    ----------
    [1] Liu R., Liu Y., Zeng S., et al. "Towards gradient-based bilevel optimization with non-convex followers and beyond," in NeurIPS, 2021.

    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        lower_loop: int,
        solver_config: Dict[str, Any],
    ):
        super(DI, self).__init__(
            ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
        )

    def optimize(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        auxiliary_opt: DifferentiableOptimizer,
        current_iter: int,
        next_operation: str = None,
        **kwargs
    ):
        """
        Executes the lower-level optimization procedure using the provided data and models.

        Parameters
        ----------
        ll_feed_dict : Dict[str, Any]
            Dictionary containing the lower-level data used for optimization. Typically includes:
            - "data" : The input data for lower-level optimization.
            - "target" : The target output (optional, depending on the task).

        ul_feed_dict : Dict[str, Any]
            Dictionary containing the upper-level data used for optimization. Typically includes:
            - "data" : The input data for upper-level optimization.
            - "target" : The target output (optional, depending on the task).

        auxiliary_model : _MonkeyPatchBase
            A patched lower model wrapped by the `higher` library. Serves as the lower-level model
            for optimization in a differentiable way.

        auxiliary_opt : DifferentiableOptimizer
            A patched optimizer for the lower-level model, wrapped by the `higher` library.
            Allows for differentiable optimization steps.

        current_iter : int
            The current iteration number of the optimization process.

        next_operation : str
            Specifies the next operation to execute during the optimization process.
            Must not be None.

        **kwargs : dict
            Additional arguments passed to the optimization procedure.

        Returns
        -------
        Dict
            A dictionary containing the input parameters and any additional keyword arguments.

        Raises
        ------
        AssertionError
            If `next_operation` is not defined.

        Notes
        -----
        Ensure that `next_operation` is defined before calling this function to specify the
        next operation in the optimization pipeline.
        """
        assert next_operation is not None, "Next operation should be defined."
        return {
            "ll_feed_dict": ll_feed_dict,
            "ul_feed_dict": ul_feed_dict,
            "auxiliary_model": auxiliary_model,
            "auxiliary_opt": auxiliary_opt,
            "current_iter": current_iter,
            **kwargs,
        }



from boat_jit.utils.op_utils import (
    grad_unused_zero,
    require_model_grad,
    update_tensor_grads,
    manual_update,
)
import jittor as jit
from jittor import Module
import copy
from typing import Dict, Any, Callable, List

from boat_jit.operation_registry import register_class
from boat_jit.dynamic_ol.dynamical_system import DynamicalSystem


@register_class
class PGDM(DynamicalSystem):
    """
    Implements the optimization procedure of Penalty-based Gradient Descent Method (PGDM) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective of the BLO problem.
    ul_objective : Callable
        The upper-level objective of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        The list of lower-level variables of the BLO problem.
    ul_var : List[jittor.Var]
        The list of upper-level variables of the BLO problem.
    lower_loop : int
        Number of iterations for lower-level optimization.
    solver_config : Dict[str, Any]
        A dictionary containing solver configurations. Expected keys include:

        - "lower_level_opt": The optimizer for the lower-level model.
        - "PGDM" (Dict): A dictionary containing the following keys:
            - "y_hat_lr": Learning rate for optimizing the surrogate variable `y_hat`.
            - "gamma_init": Initial value of the hyperparameter `gamma`.
            - "gamma_max": Maximum value of the hyperparameter `gamma`.
            - "gamma_argmax_step": Step size of the hyperparameter `gamma`.


    References
    ----------
    [1] Shen H, Chen T. "On penalty-based bilevel gradient descent method," in ICML, 2023.
    """

    def __init__(
        self,
        ll_objective: Callable,
        lower_loop: int,
        ul_model: Module,
        ul_objective: Callable,
        ll_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict[str, Any],
    ):
        super(PGDM, self).__init__(
            ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
        )
        self.ll_opt = solver_config["lower_level_opt"]
        self.ll_var = ll_var
        self.ul_var = ul_var
        self.y_hat_lr = float(solver_config["PGDM"]["y_hat_lr"])
        self.gamma_init = solver_config["PGDM"]["gamma_init"]
        self.gamma_max = solver_config["PGDM"]["gamma_max"]
        self.gamma_argmax_step = solver_config["PGDM"]["gamma_argmax_step"]
        self.gam = self.gamma_init
        self.device = solver_config["device"]

    def optimize(self, ll_feed_dict: Dict, ul_feed_dict: Dict, current_iter: int):
        """
        Execute the optimization procedure with the data from feed_dict.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            It typically includes training data, targets, and other information required to compute the LL objective.
        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            It typically includes validation data, targets, and other information required to compute the UL objective.
        current_iter : int
            The current iteration number of the optimization process.

        Returns
        -------
        Dict
            A dictionary containing the upper-level objective and the status of hypergradient computation.
        """
        y_hat = copy.deepcopy(self.ll_model)
        y_hat_opt = jit.optim.SGD(list(y_hat.parameters()), lr=self.y_hat_lr)

        if self.gamma_init > self.gamma_max:
            self.gamma_max = self.gamma_init
            print(
                "Initial gamma is larger than max gamma, proceeding with gamma_max=gamma_init."
            )
        step_gam = (self.gamma_max - self.gamma_init) / self.gamma_argmax_step
        lr_decay = min(1 / (self.gam + 1e-8), 1)
        require_model_grad(y_hat)
        for y_itr in range(self.lower_loop):
            tr_loss = self.ll_objective(ll_feed_dict, self.ul_model, y_hat)
            grads_hat = grad_unused_zero(tr_loss, y_hat.parameters())
            update_tensor_grads(list(y_hat.parameters()), grads_hat)
            manual_update(y_hat_opt, list(y_hat.parameters()))

        F_y = self.ul_objective(ul_feed_dict, self.ul_model, self.ll_model)
        loss = lr_decay * (
            F_y
            + self.gam
            * (
                self.ll_objective(ll_feed_dict, self.ul_model, self.ll_model)
                - self.ll_objective(ll_feed_dict, self.ul_model, y_hat)
            )
        )
        grads_lower = grad_unused_zero(loss, self.ll_var)
        update_tensor_grads(self.ll_var, grads_lower)
        grads_upper = grad_unused_zero(loss, self.ul_var)
        update_tensor_grads(self.ul_var, grads_upper)
        self.gam += step_gam
        self.gam = min(self.gamma_max, self.gam)
        manual_update(self.ll_opt, list(self.ll_var))
        return {"upper_loss": F_y.item()}



import jittor as jit
from jittor import Module
from typing import List, Callable, Dict
from ..higher_jit.patch import _MonkeyPatchBase
from boat_jit.utils.op_utils import update_tensor_grads

from boat_jit.operation_registry import register_class
from boat_jit.hyper_ol.hyper_gradient import HyperGradient


@register_class
class RAD(HyperGradient):
    """
    Computes the hyper-gradient of the upper-level variables using Reverse Auto Differentiation (RAD) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        List of variables optimized with the lower-level objective.
    ul_var : List[jittor.Var]
        List of variables optimized with the upper-level objective.
    solver_config : Dict[str, Any]
        Dictionary containing solver configurations, including optional dynamic operation settings.

    References
    ----------
    [1] Franceschi, Luca, et al. "Forward and reverse gradient-based hyperparameter optimization." in ICML, 2017.
    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict,
    ):
        super(RAD, self).__init__(
            ll_objective,
            ul_objective,
            ul_model,
            ll_model,
            ll_var,
            ul_var,
            solver_config,
        )
        self.dynamic_initialization = "DI" in solver_config["dynamic_op"]

    def compute_gradients(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        max_loss_iter: int = 0,
        next_operation: str = None,
        **kwargs
    ):
        """
        Compute the hyper-gradients of the upper-level variables using the provided data and patched models.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            Typically includes training data, targets, and other information required to compute the lower-level objective.

        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            Typically includes validation data, targets, and other information required to compute the upper-level objective.

        auxiliary_model : _MonkeyPatchBase
            A patched lower model wrapped by the `higher` library. It serves as the lower-level model for optimization.

        max_loss_iter : int, optional
            The number of iterations used for backpropagation. Default is 0.

        next_operation : str, optional
            The next operator for the calculation of the hypergradient. Default is None.

        **kwargs : dict
            Additional keyword arguments passed to the method.


        Returns
        -------
        Dict
            A dictionary containing the upper-level objective and the status of hypergradient computation.
        """
        assert next_operation is None, "RAD does not support any further operations."
        lower_model_params = kwargs.get(
            "lower_model_params", list(auxiliary_model.parameters())
        )
        upper_loss = self.ul_objective(
            ul_feed_dict, self.ul_model, auxiliary_model, params=lower_model_params
        )
        grads_upper = jit.grad(
            upper_loss, self.ul_var, retain_graph=self.dynamic_initialization
        )
        update_tensor_grads(self.ul_var, grads_upper)

        if self.dynamic_initialization:
            grads_lower = jit.grad(upper_loss, list(auxiliary_model.parameters(time=0)))
            update_tensor_grads(self.ll_var, grads_lower)

        return {"upper_loss": upper_loss.item(), "hyper_gradient_finished": True}



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for making ``torch.nn.Module`` subclass instances stateless."""

import abc as _abc
from collections import OrderedDict as _OrderedDict
from contextlib import contextmanager as _contextmanager
import typing as _typing
import weakref as _weakref
import warnings as _warnings

import jittor
import jittor as jit

from . import utils as _utils

# ==============================================================================
# Helper functions and attributes for MonkeyPatch modules.
# ==============================================================================

_internal_attrs = {
    
    "_backend",
    "_parameters",
    "_buffers",
    "_backward_hooks",
    "_forward_hooks",
    "_forward_pre_hooks",
    "_state_dict_hooks",
    "_load_state_dict_pre_hooks",
    "_modules",
}

# _internal_attrs = {
    
#     "_backend",
#     "_backward_hooks",
#     "_forward_hooks",
#     "_forward_pre_hooks",
#     "_state_dict_hooks",
#     "_load_state_dict_pre_hooks",
# }


_BufferType = _typing.Dict[str, _typing.Optional[jit.Var]]


@_contextmanager
def _modify_internally(fmodule):
    fmodule._being_modified_internally = True
    yield
    fmodule._being_modified_internally = False


def _patched_parameters(
    self, recurse: bool = True, time: _typing.Optional[int] = None
) -> _typing.Iterable[jit.Var]:
    r"""Returns an iterator over monkey patched module fast parameters.

    Args:
        recurse (bool): if True, then yields fast parameters of this module
            and all submodules. Otherwise, this *still* yields parameters of
            this module and all submodules, and raises a warning. This keyword
            exists only to satisfy API compatibility with
            ``torch.nn.Module.parameters``.
        time (int or None): if None, the most recent fast parameters are
            provided. The int provided stands for the number of steps since the
            module was created. *Note* that the step counter is incremented
            every time parameters are updated, so this may not align with number
            of training or evaluations steps.

    Yields:
        Parameter: module fast weights.
    """
    if getattr(self, "_fast_params", None) is None:
        raise Exception(
            "Tried to get fast weights of a monkey patched module which does "
            "not encapsulate fast weights."
        )

    if not recurse:
        _warnings.warn(
            "Calling parameters with recurse=False on a monkey patched module "
            "still returns all the fast weights of of nested patched modules."
        )

    time = -1 if time is None else time

    if not self.track_higher_grads and time not in (-1, 0):
        raise ValueError(
            "The patched model is not tracking higher gradients. Only the "
            "latest parameters are available."
        )

    return iter(self._fast_params[time])


class _MonkeyPatchBase(_abc.ABC, jit.Module):
    @_abc.abstractmethod
    def __init__(self) -> None:
        self._param_mapping: _typing.List[int] = []
        self._being_modified_internally: bool = True
        self._track_higher_grads: bool = True

    def forward(self):
        raise NotImplementedError(
            "The monkey-patching logic has failed to override self.forward "
            "on the new module, or you tried calling forward on a patched "
            "version of a module which doesn't have forward (e.g. ModuleList)."
        )

    def _expand_params(self, params: _typing.List[jit.Var]) -> _typing.List[jit.Var]:
        expanded = []
        for index in self._param_mapping:
            expanded.append(params[index])
        return expanded

    @property
    def init_fast_params(self):
        if not self.track_higher_grads:
            raise Exception(
                "Cannot get initial parameters when not tracking higher " "gradients."
            )
        return self._fast_params[0]

    @property
    def fast_params(self):
        return None if self._fast_params is None else self._fast_params[-1]

    @fast_params.setter
    def fast_params(self, value):
        value = list(value)
        if self._fast_params is None:
            self._fast_params = []
        if self.track_higher_grads:
            self._fast_params.append(value)
        else:
            self._fast_params[0] = value

    @property
    def track_higher_grads(self):
        return self._track_higher_grads

    @track_higher_grads.setter
    def track_higher_grads(self, value):
        if not isinstance(value, bool):
            raise ValueError("Expected boolean argument. Got: {}.".format(type(value)))
        self._track_higher_grads = value


def buffer_sync(
    module: jit.Module, fmodule: _MonkeyPatchBase, device: _typing.Optional[str] = None
) -> None:
    r"""One off sync (copy) of buffers in ``fmodule`` with those from ``module``."""
    for key, value in module._buffers.items():
        if not isinstance(value, jit.Var):
            fmodule._buffers[key] = value
        elif device is None:
            fmodule._buffers[key] = value.clone().detach()
        else:
            fmodule._buffers[key] = value.clone().detach().to(device)
    for name, child in module._modules.items():
        if name in fmodule._modules:
            buffer_sync(child, fmodule._modules[name], device)
        else:
            raise KeyError(
                "Did not find expected submodule {} of monkey-patched module {}.".format(name, fmodule)
            )

# from collections import deque

# def buffer_sync(module, fmodule, device=None):
#     """One-off sync (copy) of buffers in `fmodule` with those from `module`."""
#     queue = deque([(module, fmodule)])
    
#     while queue:
#         current_module, current_fmodule = queue.popleft()
        
#         # Sync buffers
#         for key, value in getattr(current_module, "_buffers", {}).items():
#             if not isinstance(value, jit.Var):
#                 current_fmodule._buffers[key] = value
#             elif device is None:
#                 current_fmodule._buffers[key] = value.clone().detach()
#             else:
#                 current_fmodule._buffers[key] = value.clone().detach()  # Jittor 没有设备概念

#         # Add child modules to the queue
#         current_modules = getattr(current_module, "_modules", {})
#         current_fmodules = getattr(current_fmodule, "_modules", {})
        
#         for name, child in current_modules.items():
#             if name in current_fmodules:
#                 queue.append((child, current_fmodules[name]))
#             else:
#                 print("fine")
#                 # raise KeyError(
#                 #     f"Did not find expected submodule {name} "
#                 #     f"of monkey-patched module {current_fmodule}."
#                 # )



# def buffer_sync(
#     module: jit.Module, fmodule: _MonkeyPatchBase, device: _typing.Optional[str] = None
# ) -> None:
#     """Synchronize buffers from `module` to `fmodule`."""
#     # 获取 module 和 fmodule 的 buffer
#     module_buffers = getattr(module, "_buffers", None)
#     fmodule_buffers = getattr(fmodule, "_buffers", None)

#     if module_buffers is None or fmodule_buffers is None:
#         raise ValueError("Failed to access _buffers for module or fmodule.")

#     # 同步 buffer
#     for key, value in module_buffers.items():
#         if not isinstance(value, jit.Var):
#             fmodule_buffers[key] = value
#         elif device is None:
#             fmodule.register_buffer(key, value.clone().detach())
#         else:
#             fmodule.register_buffer(key, value.clone().detach().to(device))

#     # 递归同步子模块
#     module_modules = getattr(module, "_modules", None)
#     fmodule_modules = getattr(fmodule, "_modules", None)

#     if module_modules is None or fmodule_modules is None:
#         raise ValueError("Failed to access _modules for module or fmodule.")

#     for name, child in module_modules.items():
#         if name in fmodule_modules:
#             buffer_sync(child, fmodule_modules[name], device)
#         else:
#             raise KeyError(
#                 f"Did not find expected submodule {name} "
#                 f"of monkey-patched module {fmodule}."
#             )





# ==============================================================================
# Helper class to use instead of actual torch.nn.Parameters when patching.
# ==============================================================================


class _ParameterPlaceholder:
    def __init__(self, name: str) -> None:
        self._param_name = name

    def __repr__(self) -> str:
        return 'Parameter placeholder ("{}")'.format(self._param_name)


_ParameterPlaceholder.__name__ = "ParameterPlaceholder"
_ParameterPlaceholder.__qualname__ = "ParameterPlaceholder"

# ==============================================================================
# Helper function for recursively patching submodules.
# ==============================================================================


def _make_functional(
    module: jit.Module,
    params_box: _typing.Sequence[_typing.Optional[_typing.List[jit.Var]]],
    params_offset: int,
    root_patched: _typing.Optional[_MonkeyPatchBase] = None,
) -> _typing.Tuple[int, _MonkeyPatchBase, _typing.Type[_MonkeyPatchBase]]:
    if isinstance(module, _MonkeyPatchBase):
        raise ValueError(
            "Monkey-patching monkey-patched modules is untested uncharted "
            "territory, so we're going to assume it's done in error. If you "
            "are doing this intentionally and need this to be supported, "
            "contact the developers of this library."
        )

    # param_names = list(
    #     name
    #     for name in module._parameters.keys()
    #     if module._parameters[name] is not None
    # )
    param_names = list(
        name
        for name, param in module.named_parameters(False)
        if param is not None
    )
    
    # print('params_name',param_names)
    _ModuleType: _typing.Type[jit.Module] = module.__class__

    # type checking of next line disabled as mypy is iffy with dynamic types
    class MonkeyPatched(_ModuleType, _MonkeyPatchBase):  # type: ignore
        _wrapped_name = type(module).__name__

        def __init__(self, original_params, root) -> None:
            # init_params = {
            # "C": module.__dict__.get("_C", None),
            # "num_classes": module.__dict__.get("_num_classes", None),
            # "layers": module.__dict__.get("_layers", None),
            # "criterion": module.__dict__.get("_criterion", None),
            # }

            # init_params = {
            # "C": module.__dict__.get("_C", None),
            # "num_classes": module.__dict__.get("_num_classes", None),
            # "layers": module.__dict__.get("_layers", None),
            # "criterion": module.__dict__.get("_criterion", None),
            # }

            # # 确保所有必要的参数都存在
            # if None in init_params.values():
            #     missing = [k for k, v in init_params.items() if v is None]
            #     raise ValueError(f"Missing initialization parameters: {missing}")
            # 手动初始化 Jittor Module 的关键属性
            # self._buffers = {}
            # self._modules = {}
            # self._parameters = {}

            # 初始化 jit.Module
            # jit.Module.__init__(self, **init_params)
            jit.Module.__init__(self)
            _MonkeyPatchBase.__init__(self)
            self._root_ref = _weakref.ref(root) if root else None

            self._fast_params = None
            self._param_names = param_names
            # print("_param_names", self._param_names)
            self._original_params = original_params
            # print("_original_params", self._original_params)
            # for pretty printing
            # self._parameters_not_recur = _OrderedDict(
            #     (name, _ParameterPlaceholder(name))
            #     for name in self._param_names
            # )
            # print(self._param_names)
            # self.load_parameters(_OrderedDict(
            #     (name, _ParameterPlaceholder(name))
            #     for name in self._param_names
            # ))
            #     for name in self._param_names)
            # for name in self._param_names:
            #     placeholder = _ParameterPlaceholder(name)
            #     # setattr(self, name, placeholder)
            #     self.safe_setattr(name, placeholder)

        @property
        def direct_submodule_call(self):
            return params_box[0] is None

        @property
        def is_root(self):
            return self._root_ref is None

        @property
        def root(self):
            if self.is_root:
                return self
            else:
                return self._root_ref()

        def safe_setattr(self, name, value):
                """
                Safely set attributes for a Jittor Module, handling Jittor's dynamic properties
                like _modules, _parameters, and _buffers without causing recursion issues.
                """
                def remove_from(*dicts):
                    for d in dicts:
                        if name in d:
                            del d[name]

                # Special handling for _parameters
                # params = getattr(self, "_parameters", None)
                params = self.named_parameters(False)
                # print('params',params)
                if params is not None and name in params:
                    if not isinstance(value, jit.Var):
                        raise TypeError(
                            "Require Tensor as fast weights. "
                            "Got {}".format(type(value).__name__)
                        )

                    if not self._being_modified_internally:
                        # Additional behaviour for when fast weights are being directly modified
                        old_value = self._parameters[name]
                        fast_params = self.root.fast_params[:]
                        if not fast_params:
                            raise Exception(
                                "Cannot assign parameters to patched module which "
                                "does not have implicit fast parameters."
                            )
                        replacement_index = _utils._find_param_in_list(old_value, fast_params)
                        fast_params[replacement_index] = value
                        self.update_params(fast_params)

                    # Change parameters in place, usually during boxed_forward pass
                    # self._parameters[name] = value
                    # setattr(self,name,value)
                    object.__setattr__(self, name, value)
                else:
                    # Special handling for _modules
                    modules = getattr(self, "_modules", None)
                    if isinstance(value, jit.Module):
                        if modules is None:
                            raise AttributeError(
                                "cannot assign module before Module.__init__() call"
                            )
                        remove_from(self.__dict__, self._parameters, self._buffers)
                        # modules[name] = valu
                        self.add_module(name,value)
                    elif modules is not None and name in modules:
                        if value is not None:
                            raise TypeError(
                                (
                                    "cannot assign '{}' as child module '{}'"
                                    "(jit.Module or None expected)"
                                ).format(type(value).__name__, name)
                            )
                        self.add_module(name,value)
                    else:
                        # Special handling for _buffers
                        buffers = getattr(self, "_buffers", None)
                        if buffers is not None and name in buffers:
                            if value is not None and not isinstance(value, jit.Var):
                                raise TypeError(
                                    "cannot assign '{}' as buffer '{}' "
                                    "(jit.Var or None expected)".format(type(value).__name__, name)
                                )
                            # buffers[name] = value
                            self.register_buffer(name, value)
                        else:
                            # Default behavior
                            object.__setattr__(self, name, value)


        # def __setattr__(self, name, value):
        #     def remove_from(*dicts):
        #         for d in dicts:
        #             if name in d:
        #                 del d[name]

        #     # print(f"Attempting to set attribute: {name}, value: {value}")
        #     # if hasattr(self, name):
        #     #     existing_value = getattr(self, name, None)
        #     #     print(f"Existing attribute '{name}': {existing_value}")
        #     # else:
        #     #     print(f"Attribute '{name}' does not exist yet.")
            
        #     # 特殊处理 Jittor 的内部属性
        #     # if name in {"_modules", "_parameters", "_buffers"}:
        #     #     try:
        #     #         object.__setattr__(self, name, value)
        #     #         print(f"Attribute '{name}' set successfully.")
        #     #     except Exception as e:
        #     #         print(f"Failed to set attribute '{name}': {e}")
        #     #     return

        #     params = self.__dict__.get("_parameters")
        #     print('params',params)
        #     if params is not None and name in params:
        #         if not isinstance(value, jit.Var):
        #             raise TypeError(
        #                 "Require Tensor as fast weights. "
        #                 "Got {}".format(type(value).__name__)
        #             )

        #         if not self._being_modified_internally:
        #             # Additional behaviour for when fast weights are being
        #             # directly modified goes here:
        #             old_value = self._parameters[name]
        #             fast_params = self.root.fast_params[:]
        #             if not fast_params:
        #                 raise Exception(
        #                     "Cannot assign parameters to patched module which "
        #                     "does not have implicit fast parameters."
        #                 )
        #             replacement_index = _utils._find_param_in_list(
        #                 old_value, fast_params
        #             )
        #             fast_params[replacement_index] = value
        #             self.update_params(fast_params)

        #         # Change parameters in place, usually during boxed_forward pass
        #         self._parameters[name] = value
        #     else:
        #         # modules = self.__dict__.get("_modules")
        #         modules = getattr(self, "_modules", None)
        #         print("modules", modules)
        #         if isinstance(value, jit.Module):
        #             # print(value)
        #             if modules is None:
        #                 raise AttributeError(
        #                     "cannot assign module before Module.__init__() " "call"
        #                 )
        #             remove_from(self.__dict__, self._parameters, self._buffers)
        #             modules[name] = value
        #             # self.add_module(name, value)
        #         elif modules is not None and name in modules:
        #             if value is not None:
        #                 raise TypeError(
        #                     (
        #                         "cannot assign '{}' "
        #                         "as child module '{}'"
        #                         "(torch.nn.Module or None expected)"
        #                     ).format(type(value).__name__, name)
        #                 )
        #             modules[name] = value
        #         else:
        #             buffers = self.__dict__.get("_buffers")
        #             if buffers is not None and name in buffers:
        #                 if value is not None and not isinstance(value, jit.Var):
        #                     raise TypeError(
        #                         "cannot assign '{}' as buffer '{}' "
        #                         "(torch.Tensor or None expected)".format(
        #                             type(value).__name__, name
        #                         )
        #                     )
        #                 buffers[name] = value
        #             else:
        #                 object.__setattr__(self, name, value)


    MonkeyPatched.__name__ = "InnerFunctional" + type(module).__name__
    MonkeyPatched.__qualname__ = MonkeyPatched.__name__

    fmodule = MonkeyPatched(module.parameters(), root=root_patched)

    # If a root module hasn't been defined yet, this fmodule is the root
    if not root_patched:
        root_patched = fmodule

    # use 1 as dummy list item since we are only counting
    # num_params = len([1 for p in module._parameters.values() if p is not None])
    num_params = len([1 for p,_ in module.named_parameters(False) if _ is not None])
    # print('num_params',num_params)
    # # Copy over all attributes
    # print(fmodule.__dict__.keys())
    # print(fmodule._modules.items())
    for name, attr in module.__dict__.items():
        # print('set_module',name)
        if name in _internal_attrs:
            continue
        # setattr(fmodule, name, attr)
        fmodule.safe_setattr(name, attr)
    # k=0
    # # print(module['_param_mapping'])
    # for name, attr in module.__dict__.items():
    #     k+=1
    #     if name in _internal_attrs:
    #         continue

    #     # setattr(fmodule, name, attr)
    #     fmodule.safe_setattr(name, attr)
    # print(len(module._parameters.keys()))
    # print('new!!!!!!!!!!!!')
    # for k,b  in enumerate(module.parameters()):
    #     print('items ',k ,b.shape)

    # Deal with "None"-style params
    
    with _modify_internally(fmodule):
        # for name, attr in module.__dict__['_parameters'].items():
        for name, attr in module.named_parameters(False):
            if isinstance(attr, jit.Var) and not attr.stop_grad:
                continue
            else:
                # setattr(fmodule, name, attr)
                fmodule.safe_setattr(name, attr)

    child_params_offset = params_offset + num_params
    # print(child_params_offset)
    # k=1
    # for name, child in module._modules.items():
    # for name, child in enumerate(module.modules()):
    #     print(child)
    for name, child in module._modules.items():
        # k+=1
        # print('_make_functional',name)
        # print(type(child).__name__)
        # print('old!!!!!!!!!!!!',k)
        # if name =='classifier':
        #     print(child.__dict__)
        #     print(child.weight.shape)
        # if 'cell' in name:
        #     print('cells test')
        #     print(child._modules.keys())
        # if hasattr(child,'shape'):
        #     print("name",name,child.shape)

        child_params_offset, fchild, _ = _make_functional(
            child, params_box, child_params_offset, root_patched
        )
        # fmodule._modules[name] = fchild
        # setattr(fmodule, name, fchild)
        fmodule.safe_setattr(name, fchild)

    true_forward = getattr(type(module), "execute", None) or getattr(
        type(module), "forward", None
    )
    # true_forward = type(module).forward

    def patched_forward(self, *args, params=None, **kwargs):
        # print('run patched_forward!!!')
        if self.direct_submodule_call:
            # If submodule was called directly, run intialisation that happens
            # at top level call. If *full set of params* is provided here, it
            # will use those. If not, it will fall back on fast weights.
            # In the future, we should be able to support passing only the
            # submodule (+ children) weights here, but that's not simple.
            self.root._refill_params_box(params)

        with _modify_internally(self):
            for name, param in zip(
                self._param_names,
                params_box[0][params_offset : params_offset + num_params],
            ):
                setattr(self, name, param)

            # This snippet deals with torch.nn.{RNN,GRU,LSTM}
            if hasattr(self, "_flat_weights_names"):
                self._flat_weights = [
                    self._parameters[wn] for wn in self._flat_weights_names
                ]

        # Call true_forward after some checks
        with _warnings.catch_warnings():
            # If running RNNs on GPU, surpress the warnings due to flattening
            # not happening here. Maybe we should raise a warning of our own?
            is_RNN = isinstance(module, jit.nn.RNN)
            if is_RNN and jit.flags.use_cuda:
                _warnings.simplefilter("ignore", category=UserWarning)

            return true_forward(self, *args, **kwargs)

    # setattr(MonkeyPatched, "forward", patched_forward)
    setattr(MonkeyPatched, "execute", patched_forward)
    # MonkeyPatched.safe_setattr("execute", patched_forward)

    def flatten_parameters(self):
        return  # no-op

    # This (hopefully) avoids trouble on GPU with torch.nn.{RNN,GRU,LSTM}
    if hasattr(module, "flatten_parameters"):
        setattr(MonkeyPatched, "flatten_parameters", flatten_parameters)
        # MonkeyPatched.safe_setattr("flatten_parameters", flatten_parameters)
    return child_params_offset, fmodule, type(fmodule)


def _update_patched_params(
    fmodule: _MonkeyPatchBase,
    params_box: _typing.Sequence[_typing.List[jit.Var]],
    params_offset: int,
) -> int:
    # num_params = len([1 for p in fmodule._parameters.values() if p is not None])
    num_params = len([1 for p,_ in fmodule.named_parameters(False) if _ is not None])
    child_params_offset = params_offset + num_params
    for name, child in fmodule._modules.items():
        child_params_offset = _update_patched_params(
            child, params_box, child_params_offset
        )

    with _modify_internally(fmodule):
        for name, param in zip(
            fmodule._param_names,
            params_box[0][params_offset : params_offset + num_params],
        ):
            setattr(fmodule, name, param)
    return child_params_offset


# ==============================================================================
# The main function which does the monkey patching.
# ==============================================================================
_EncapsulatorType = _typing.Optional[
    _typing.Callable[[_MonkeyPatchBase, jit.Var], None]
]


def make_functional(
    module: jit.Module, encapsulator: _EncapsulatorType = None
) -> _MonkeyPatchBase:
    r"""Returns a stateless version of an ``nn.Module`` instance."""
    params_box = [None]
    _, fmodule, MonkeyPatched = _make_functional(module, params_box, 0)

    top_name = "Functional" + MonkeyPatched._wrapped_name
    MonkeyPatched.__name__ = MonkeyPatched.__qualname__ = top_name
    ###################
    # MonkeyPatched.boxed_forward = MonkeyPatched.forward
    MonkeyPatched.boxed_forward = MonkeyPatched.execute
    #############
    param_mapping = _utils._get_param_mapping(module, [], [])
    setattr(fmodule, "_param_mapping", param_mapping)

    def _refill_params_box(self, params):
        if params is not None:
            self.fast_params = params  # update view on latest fast params
        elif self.fast_params is None:
            raise ValueError(
                "params keyword must be provided if patched module not "
                "tracking its own fast parameters"
            )

        # Copy fast parameters into params_box for use in boxed_forward
        params_box[0] = self._expand_params(self.fast_params)

    def _patched_forward(self, *args, params=None, **kwargs):
        self._refill_params_box(params)

        output = self.boxed_forward(*args, **kwargs)

        # Clean up
        params_box[0] = None

        return output

    def _update_params(self, params):
        self.fast_params = params
        # print(params)
        # print('lang')
        # print('right',params[-1].shape)
        # print('right',params[-2].shape)
        # print(params[-3].shape)
        # print(params[-4].shape)
        # print(len(params))
        params = self._expand_params(params)
        # print(len(params))
        # print('not right',params[-1].shape)
        # print('not right',params[-2].shape)
        _update_patched_params(self, [params], 0)

    ##############
    setattr(MonkeyPatched, "execute", _patched_forward)
    ##############
    # setattr(MonkeyPatched, "forward", _patched_forward)
    setattr(MonkeyPatched, "parameters", _patched_parameters)
    setattr(MonkeyPatched, "update_params", _update_params)
    setattr(MonkeyPatched, "_refill_params_box", _refill_params_box)
    
    if encapsulator is not None:
        encapsulator(fmodule, module)
    return fmodule


# ==============================================================================
# Convenience functions and decorators for hiding away a lot of the complexity
# of creating patched modules, taking their parameters, and linking patched
# modules to a differentiable optimizer.
# ==============================================================================


def monkeypatch(
    module: jit.Module,
    device: _typing.Optional[str] = None,
    copy_initial_weights: bool = True,
    track_higher_grads: bool = True,
) -> _MonkeyPatchBase:
    r"""Create a monkey-patched stateless version of a module.

    This function produces a monkey-patched version of a module, and returns a
    copy of its parameters for use as fast weights. Where the original module
    or any of its submodules have state (e.g. batch norm), this will be copied
    too, but further updates (e.g. during inner loop training) will cause these
    to diverge without changing the state of the original module.

    Args:
        module: a ``torch.nn.Module`` subclass instance.
        device (optional): a device to cast the fast weights and state to.
        copy_initial_weights: if True, the weights of the patched module are
            copied to form the initial weights of the patched module, and thus
            are not part of the gradient tape when unrolling the patched module.
            If this is set to False, the actual module weights will be the
            initial weights of the patched module. This is useful when doing
            MAML, for example.
        track_higher_grads: if True, during unrolled optimization the graph be
            retained, and the fast weights will bear grad funcs, so as to permit
            backpropagation through the optimization process. Setting this to
            False allows ``monkeypatch`` to be used in "test mode", without
            potentially tracking higher order gradients. This can be useful when
            running the training loop at test time, e.g. in k-shot learning
            experiments, without incurring a significant memory overhead.

    Returns:
        ``fmodule``: a "stateless" version of the original module, for which calls
        to forward take the additional kwarg-only parameter ``params``, which
        should be a list of torch tensors requiring gradients, ideally
        provided by this function (see below) or by an update step from one
        of the optimizers in ``higher.optim``.
    """

    def encapsulator(fmodule: _MonkeyPatchBase, module: jit.Module) -> None:
        if copy_initial_weights:
            params = _utils.get_func_params(module, device=device)
        else:   
            params = [
                p.clone() if device is None else p.clone().to(device)
                for p in module.parameters()
            ]
        # params = [
        #     p.clone() if device is None else p.clone().to(device)
        #     for p in module.parameters()
        # ]
        # print('now')
        # print(len(params))
        # print(params[-2].shape)
        # print(params[-1].shape)
        # print('test',fmodule.classifier.weight.shape)
        # print('test',module.classifier.weight.shape)
        # print('test',len(fmodule.cells))
        # print('test',len(module.cells))
        buffer_sync(module, fmodule, device)
        # print('test',len(fmodule.cells))
        # print('test',len(module.cells))
        # print('test',fmodule.classifier.weight.shape)
        # print('test',module.classifier.weight.shape)
        fmodule.update_params(params)
        # new_params= [
        #         p.clone() if device is None else p.clone().to(device)
        #         for p in fmodule.parameters()
        #     ]
        # print(new_params[-2].shape)
        # print(new_params[-1].shape)
        # print(len(new_params))
        # print('test',fmodule.classifier.weight.shape)
        # print('test',module.classifier.weight.shape)

    fmodule = make_functional(module, encapsulator=encapsulator)
    # print('final_test',module.classifier.weight.shape)
    fmodule.track_higher_grads = track_higher_grads

    return fmodule



import jittor as jit
from jittor import Module
from typing import List, Callable, Dict
from boat_jit.higher_jit.patch import _MonkeyPatchBase
from boat_jit.utils.op_utils import update_tensor_grads

from boat_jit.operation_registry import register_class
from boat_jit.hyper_ol.hyper_gradient import HyperGradient


@register_class
class IAD(HyperGradient):
    """
    Computes the hyper-gradient of the upper-level variables using Initialization-based Auto Differentiation (IAD) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        List of variables optimized with the lower-level objective.
    ul_var : List[jittor.Var]
        List of variables optimized with the upper-level objective.
    solver_config : Dict[str, Any]
        Dictionary containing solver configurations.

    References
    ----------
    [1] Finn C., Abbeel P., Levine S., "Model-agnostic meta-learning for fast adaptation of deep networks", in ICML, 2017.
    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict,
    ):
        super(IAD, self).__init__(
            ll_objective,
            ul_objective,
            ul_model,
            ll_model,
            ll_var,
            ul_var,
            solver_config,
        )

    def compute_gradients(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        max_loss_iter: int = 0,
        hyper_gradient_finished: bool = False,
        next_operation: str = None,
        **kwargs
    ):
        """
        Compute the hyper-gradients of the upper-level variables with the data from feed_dict and patched models.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization. It typically includes training data, targets, and other information required to compute the LL objective.
        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization. It typically includes validation data, targets, and other information required to compute the UL objective.
        auxiliary_model : _MonkeyPatchBase
            A patched lower model wrapped by the `higher` library. It serves as the lower-level model for optimization.
        max_loss_iter : int
            The number of iterations used for backpropagation.
        next_operation : str
            The next operator for the calculation of the hypergradient.
        hyper_gradient_finished : bool
            A boolean flag indicating whether the hypergradient computation is finished.

        Returns
        -------
        Dict
            A dictionary containing the upper-level objective and the status of hypergradient computation.
        """

        if next_operation is not None:
            lower_model_params = kwargs.get(
                "lower_model_params", list(auxiliary_model.parameters())
            )
            hparams = list(auxiliary_model.parameters(time=0))
            return {
                "ll_feed_dict": ll_feed_dict,
                "ul_feed_dict": ul_feed_dict,
                "auxiliary_model": auxiliary_model,
                "max_loss_iter": max_loss_iter,
                "hyper_gradient_finished": hyper_gradient_finished,
                "hparams": hparams,
                "lower_model_params": lower_model_params,
                **kwargs,
            }
        else:
            lower_model_params = kwargs.get(
                "lower_model_params", list(auxiliary_model.parameters())
            )
            ul_loss = self.ul_objective(
                ul_feed_dict, self.ul_model, auxiliary_model, params=lower_model_params
            )
            grads_upper = jit.grad(ul_loss, list(auxiliary_model.parameters(time=0)))
            update_tensor_grads(self.ul_var, grads_upper)
            return {"upper_loss": ul_loss.item(), "hyper_gradient_finished": True}



import jittor as jit
from jittor import Module
from typing import List, Callable, Dict
from ..higher_jit.patch import _MonkeyPatchBase
from boat_jit.utils.op_utils import update_tensor_grads

from boat_jit.operation_registry import register_class
from boat_jit.hyper_ol.hyper_gradient import HyperGradient


@register_class
class FD(HyperGradient):
    """
    Computes the hyper-gradient of the upper-level variables using Finite Differentiation (FD) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        List of variables optimized with the lower-level objective.
    ul_var : List[jittor.Var]
        List of variables optimized with the upper-level objective.
    solver_config : Dict[str, Any]
        Dictionary containing solver configurations. Expected keys include:

        - `r` (float): Perturbation radius for finite differences.
        - `lower_level_opt` (torch.optim.Optimizer): Lower-level optimizer configuration.
        - `dynamic_op` (str): Indicates dynamic initialization type (e.g., "DI").
        - GDA-specific parameters if applicable, such as:
            - `alpha_init` (float): Initial learning rate for GDA.
            - `alpha_decay` (float): Decay factor for GDA.

    Attributes
    ----------
    ll_lr : float
        Learning rate for the lower-level optimizer, extracted from `lower_level_opt`.
    dynamic_initialization : bool
        Indicates whether dynamic initialization is enabled (based on `dynamic_op`).
    _r : float
        Perturbation radius for finite differences, used for gradient computation.
    alpha : float
        Initial learning rate for GDA operations.
    alpha_decay : float
        Decay factor applied to the learning rate for GDA.
    gda_loss : Callable, optional
        Custom loss function for GDA operations, if specified in `solver_config`.

    References
    ----------
    [1] H. Liu, K. Simonyan, Y. Yang, "DARTS: Differentiable Architecture Search," in ICLR, 2019.
    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict,
    ):
        super(FD, self).__init__(
            ll_objective,
            ul_objective,
            ul_model,
            ll_model,
            ll_var,
            ul_var,
            solver_config,
        )
        self.ll_lr = solver_config["lower_level_opt"].defaults["lr"]
        self.dynamic_initialization = "DI" in solver_config["dynamic_op"]
        self._r = solver_config["FD"]["r"]
        self.alpha = solver_config["GDA"]["alpha_init"]
        self.alpha_decay = solver_config["GDA"]["alpha_decay"]
        self.gda_loss = solver_config.get("gda_loss", None)

    def compute_gradients(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        max_loss_iter: int = 0,
        hyper_gradient_finished: bool = False,
        next_operation: str = None,
        **kwargs
    ):
        """
        Compute the hyper-gradients of the upper-level variables with the data from `feed_dict` and patched models.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            It typically includes training data, targets, and other information required to compute the LL objective.

        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            It typically includes validation data, targets, and other information required to compute the UL objective.

        auxiliary_model : _MonkeyPatchBase
            A patched lower model wrapped by the `higher` library.
            It serves as the lower-level model for optimization.

        max_loss_iter : int, optional
            The number of iterations used for backpropagation. Default is 0.

        hyper_gradient_finished : bool, optional
            A boolean flag indicating whether the hyper-gradient computation is finished. Default is False.

        next_operation : str, optional
            The next operator for the calculation of the hyper-gradient. Default is None.

        Returns
        -------
        dict
            A dictionary containing:
            - "upper_loss": The current upper-level objective value.
            - "hyper_gradient_finished": A boolean indicating whether the hyper-gradient computation is complete.

        Raises
        ------
        AssertionError
            If `next_operation` is not None, as FD does not support `next_operation`.
        """
        assert next_operation is None, "FD does not support next_operation"
        import time
        start_time = time.perf_counter()
        lower_model_params = kwargs.get(
            "lower_model_params", list(auxiliary_model.parameters())
        )
        loss = self.ul_objective(
            ul_feed_dict, self.ul_model, auxiliary_model, params=lower_model_params
        )
        # print(type(loss))
        # print(type(self.ul_var))
        # loss = loss.astype('float32')  # 强制转换为 float32
        # self.ul_var = [v.astype('float32') for v in self.ul_var]
        # for variable in self.ul_var:
        #     print(type(variable))
        # import time
        print('step 1 time',time.perf_counter() - start_time)
        start_time = time.perf_counter()
        # dalpha = jit.grad(loss, self.ul_var)
        dalpha = jit.jittor_core.grad(loss,self.ul_var,retain_graph=True)

        vector = jit.grad(
            loss,
            lower_model_params,
            retain_graph=self.dynamic_initialization,
        )

        # dalpha = jit.grad(loss, self.ul_var, retain_graph=True)

        print('step 2 time',time.perf_counter() - start_time)
        start_time = time.perf_counter()


        implicit_grads = self._hessian_vector_product(
            vector, ll_feed_dict, ul_feed_dict
        )

        for g, ig in zip(dalpha, implicit_grads):
            g.update(g - ig * self.ll_lr)

        if self.dynamic_initialization:
            grads_lower = jit.grad(loss, list(auxiliary_model.parameters(time=0)))
            update_tensor_grads(self.ll_var, grads_lower)
        print('step 3 time',time.perf_counter() - start_time)


        update_tensor_grads(self.ul_var, dalpha)

        return {"upper_loss": loss.item(), "hyper_gradient_finished": True}

    def _hessian_vector_product(self, vector, ll_feed_dict, ul_feed_dict):
        """
        Compute the first-order approximation of the second-order derivative of upper-level variables.

        Parameters
        ----------
        vector : List[Tensor]
            A vector used for computing the Hessian-vector product.

        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.

        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.

        Returns
        -------
        List[Tensor]
            A list of tensors representing the first-order approximation of the second-order derivative (Hessian-vector product).

        Notes
        -----
        The method computes the Hessian-vector product using finite difference approximation, and the hyper-parameter `_r` is used for scaling the perturbation.
        """
        # Compute eta
        vector_flat = jit.concat([v.flatten() for v in vector])
        eta = self._r / vector_flat.norm()

        # Update parameters: w+ = w + eta * vector
        for p, v in zip(self.ll_var, vector):
            p.update(p + v * eta)

        # Compute loss and gradients for w+
        if self.gda_loss is not None:
            ll_feed_dict["alpha"] = self.alpha
            loss = self.gda_loss(
                ll_feed_dict, ul_feed_dict, self.ul_model, self.ll_model
            )
        else:
            loss = self.ll_objective(ll_feed_dict, self.ul_model, self.ll_model)
        grads_p = jit.grad(loss, self.ul_var)

        # Update parameters: w- = w - 2 * eta * vector
        for p, v in zip(self.ll_var, vector):
            p.update(p - 2 * eta * v)

        # Compute loss and gradients for w-
        if self.gda_loss is not None:
            loss = self.gda_loss(
                ll_feed_dict, ul_feed_dict, self.ul_model, self.ll_model
            )
        else:
            loss = self.ll_objective(ll_feed_dict, self.ul_model, self.ll_model)
        grads_n = jit.grad(loss, self.ul_var)

        # Restore parameters: w = w + eta * vector
        for p, v in zip(self.ll_var, vector):
            p.update(p + eta * v)

        # Compute Hessian-vector product approximation
        return [(gp - gn) / (2 * eta) for gp, gn in zip(grads_p, grads_n)]



from boat_jit.utils.op_utils import (
    grad_unused_zero,
    require_model_grad,
    update_tensor_grads,
    stop_model_grad,
    manual_update,
)

import jittor as jit
from jittor import Module
import copy
from typing import Dict, Any, Callable, List
from boat_jit.operation_registry import register_class
from boat_jit.dynamic_ol.dynamical_system import DynamicalSystem


@register_class
class VFM(DynamicalSystem):
    """
    Implements the optimization procedure of Value-function based First-Order Method (VFM) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        A list of lower-level variables of the BLO problem.
    ul_var : List[jittor.Var]
        A list of upper-level variables of the BLO problem.
    lower_loop : int
        The number of iterations for lower-level optimization.
    solver_config : Dict[str, Any]
        A dictionary containing configurations for the solver. Expected keys include:

        - "lower_level_opt" (torch.optim.Optimizer): Optimizer for the lower-level model.
        - "VFM" (Dict): Configuration for the VFM algorithm:
            - "y_hat_lr" (float): Learning rate for optimizing the surrogate variable `y_hat`.
            - "eta" (float): Step size for value-function updates.
            - "u1" (float): Hyperparameter controlling the penalty in the value function.
        - "device" (str): Device on which computations are performed, e.g., "cpu" or "cuda".

    References
    ----------
    [1] R. Liu, X. Liu, X. Yuan, S. Zeng and J. Zhang, "A Value-Function-based
        Interior-point Method for Non-convex Bi-level Optimization," in ICML, 2021.
    """

    def __init__(
        self,
        ll_objective: Callable,
        lower_loop: int,
        ul_model: Module,
        ul_objective: Callable,
        ll_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict[str, Any],
    ):
        super(VFM, self).__init__(
            ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
        )
        self.ll_opt = solver_config["lower_level_opt"]
        self.ll_var = ll_var
        self.ul_var = ul_var
        self.y_hat_lr = float(solver_config["VFM"]["y_hat_lr"])
        self.eta = solver_config["VFM"]["eta"]
        self.u1 = solver_config["VFM"]["u1"]
        self.device = solver_config["device"]

    def optimize(self, ll_feed_dict: Dict, ul_feed_dict: Dict, current_iter: int):
        """
        Execute the optimization procedure with the data from feed_dict.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            It typically includes training data, targets, and other information required to compute the LL objective.
        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            It typically includes validation data, targets, and other information required to compute the UL objective.
        current_iter : int
            The current iteration number of the optimization process.

        Returns
        -------
        Dict
            A dictionary containing the upper-level objective and the status of hypergradient computation.
        """
        y_hat = copy.deepcopy(self.ll_model)
        y_hat_opt = jit.optim.SGD(y_hat.parameters(), lr=self.y_hat_lr, momentum=0.9)
        n_params_y = sum([p.numel() for p in self.ll_model.parameters()])
        n_params_x = sum([p.numel() for p in self.ul_model.parameters()])
        delta_f = jit.zeros((n_params_x + n_params_y,))
        delta_F = jit.zeros((n_params_x + n_params_y,))

        def g_x_xhat_w(y, y_hat, x):
            loss = self.ll_objective(ll_feed_dict, x, y) - self.ll_objective(
                ll_feed_dict, x, y_hat
            )
            grad_y = grad_unused_zero(loss, list(y.parameters()), retain_graph=True)
            grad_x = grad_unused_zero(loss, list(x.parameters()))
            return loss, grad_y, grad_x

        require_model_grad(y_hat)
        for y_itr in range(self.lower_loop):
            # y_hat_opt.zero_grad()
            tr_loss = self.ll_objective(ll_feed_dict, self.ul_model, y_hat)
            grads_hat = grad_unused_zero(tr_loss, y_hat.parameters())
            update_tensor_grads(list(y_hat.parameters()), grads_hat)
            # y_hat_opt.step()
            manual_update(y_hat_opt, list(y_hat.parameters()))
        F_y = self.ul_objective(ul_feed_dict, self.ul_model, self.ll_model)

        grad_F_y = grad_unused_zero(
            F_y, list(self.ll_model.parameters()), retain_graph=True
        )
        grad_F_x = grad_unused_zero(F_y, list(self.ul_model.parameters()))
        stop_model_grad(y_hat)
        loss, gy, gx_minus_gx_k = g_x_xhat_w(self.ll_model, y_hat, self.ul_model)
        delta_F[:n_params_y].update(
            jit.concat([fc_param.flatten().clone() for fc_param in grad_F_y])
        )
        delta_f[:n_params_y].update(
            jit.concat([fc_param.flatten().clone() for fc_param in gy])
        )

        delta_F[n_params_y:].update(
            jit.concat([fc_param.flatten().clone() for fc_param in grad_F_x])
        )
        delta_f[n_params_y:].update(
            jit.concat([fc_param.flatten().clone() for fc_param in gx_minus_gx_k])
        )

        # Compute squared norm of delta_f
        norm_dq = (delta_f * delta_f).sum()

        # Compute dot product
        dot = (delta_F * delta_f).sum()

        scaling_factor = jit.nn.relu((self.u1 * loss - dot) / (norm_dq + 1e-8))  #
        d = delta_F + scaling_factor * delta_f

        y_grad = []
        x_grad = []
        all_numel = 0
        for _, param in enumerate(self.ll_model.parameters()):
            sliced = d[all_numel : all_numel + param.numel()]
            reshaped = sliced.reshape(tuple(param.shape))
            y_grad.append(reshaped.clone())
            all_numel += param.numel()

        for _, param in enumerate(self.ul_model.parameters()):
            sliced = d[all_numel : all_numel + param.numel()]
            reshaped = sliced.reshape(tuple(param.shape))
            x_grad.append(reshaped.clone())
            all_numel += param.numel()

        update_tensor_grads(self.ll_var, y_grad)
        update_tensor_grads(self.ul_var, x_grad)
        manual_update(self.ll_opt, list(self.ll_var))
        return {"upper_loss": F_y.item()}



# Copyright (c) Facebook, Inc. and its affiliates.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import patch  # noqa: F401
from . import optim  # noqa: F401
from . import utils  # noqa: F401

from .optim import get_diff_optim, create_diff_optim  # noqa: F401
from .optim import register_optim  # noqa: F401
from .patch import monkeypatch

from contextlib import contextmanager as _contextmanager
import typing as _typing
import jittor as jit


@_contextmanager
def innerloop_ctx(
    model: jit.Module,
    opt: jit.optim.Optimizer,
    device: _typing.Optional[str] = None,
    copy_initial_weights: bool = True,
    override: optim._OverrideType = None,
    track_higher_grads: bool = True,
):
    r"""A context manager for writing differentiable inner loops.

    Args:
        model: a ``jit.Module`` subclass instance.
        opt: an existing optimizer, assumed to be an instance of
            ``jit.optim.Optimizer``, of a supported type which is either
            defined in ``jit.optim``, or a custom implementation which has
            been added to higher at runtime by using ``higher.register_optim``.
            We assume this optimizer tracks the parameters (or some subset
            thereof) of a single ``jit.Module`` instance, with support for
            parameter groups.
        device (optional): a device to cast the fast weights and state to. If
            not specified, the device used for corresponding weights of
            ``model`` will be used.
        copy_initial_weights: if true, the weights of the patched module are
            copied to form the initial weights of the patched module, and thus
            are not part of the gradient tape when unrolling the patched module.
            If this is set to False, the actual module weights will be the
            initial weights of the patched module. This is useful when doing
            MAML, for example.
        override (optional): a dictionary mapping optimizer settings (i.e. those
            which would be passed to the optimizer constructor or provided
            within parameter groups) to either singleton lists of override
            values, or to a list of override values of length equal to the
            number of parameter groups. If a single override is provided for a
            keyword, it is used for all parameter groups. If a list is provided,
            the ``i``\ th element of the list overrides the corresponding
            setting in the ``i``\ th parameter group. This permits the passing
            of tensors requiring gradient to differentiable optimizers for use
            as optimizer settings.
        track_higher_grads: if True, during unrolled optimization the graph be
            retained, and the fast weights will bear grad funcs, so as to permit
            backpropagation through the optimization process. Setting this to
            False allows ``innerloop_ctx`` to be used in "test mode", without
            potentially tracking higher order gradients. This can be useful when
            running the training loop at test time, e.g. in k-shot learning
            experiments, without incurring a significant memory overhead.

    Yields:
        A ``(fmodule, diffopt)`` tuple. where ``fmodule`` is a "stateless"
        version of the original module, for which calls to forward take the
        additional kwarg-only parameter ``params``, which should be a list of
        jit tensors requiring gradients, ideally provided by this function
        (see below) or by an update step from one of the optimizers in
        ``higher_jit.optim``. And ``diffopt`` is an initialized
        ``DifferentiableOptimizer`` instance of the right subtype.
    """

    fmodel = monkeypatch(
        model,
        device,
        copy_initial_weights=copy_initial_weights,
        track_higher_grads=track_higher_grads,
    )
    diffopt = optim.get_diff_optim(
        opt,
        model.parameters(),
        fmodel=fmodel,
        device=device,
        override=override,
        track_higher_grads=track_higher_grads,
    )
    yield fmodel, diffopt


__all__: list = ["innerloop_ctx"]



import jittor as jit
from jittor import Module
from typing import List, Callable, Dict
from ..higher_jit.patch import _MonkeyPatchBase
from boat_jit.utils.op_utils import update_tensor_grads, conjugate_gradient

from boat_jit.operation_registry import register_class
from boat_jit.hyper_ol.hyper_gradient import HyperGradient


@register_class
class CG(HyperGradient):
    """
    Computes the hyper-gradient of the upper-level variables using Conjugate Gradient (CG) [1].

    Parameters
    ----------
    ll_objective : Callable
        The lower-level objective function of the BLO problem.
    ul_objective : Callable
        The upper-level objective function of the BLO problem.
    ll_model : jittor.Module
        The lower-level model of the BLO problem.
    ul_model : jittor.Module
        The upper-level model of the BLO problem.
    ll_var : List[jittor.Var]
        List of variables optimized with the lower-level objective.
    ul_var : List[jittor.Var]
        List of variables optimized with the upper-level objective.
    solver_config : Dict[str, Any]
        Dictionary containing solver configurations. Expected keys include:

        - `r` (float): Perturbation radius for finite differences.
        - `lower_level_opt` (torch.optim.Optimizer): Lower-level optimizer configuration.
        - `dynamic_op` (str): Indicates dynamic initialization type (e.g., "DI").
        - GDA-specific parameters if applicable, such as:
            - `alpha_init` (float): Initial learning rate for GDA.
            - `alpha_decay` (float): Decay factor for GDA.

    Attributes
    ----------
    ll_lr : float
        Learning rate for the lower-level optimizer, extracted from `lower_level_opt`.
    dynamic_initialization : bool
        Indicates whether dynamic initialization is enabled (based on `dynamic_op`).
    tolerance : float
        The tolerance for approximation.
    K : int
        Number of iterations for CG approximation.
    alpha : float
        Initial learning rate for GDA operations.
    alpha_decay : float
        Decay factor applied to the learning rate for GDA.
    gda_loss : Callable, optional
        Custom loss function for GDA operations, if specified in `solver_config`.

    References
    ----------
    [1] Pedregosa F. "Hyperparameter optimization with approximate gradient," in ICML, 2016.
    """

    def __init__(
        self,
        ll_objective: Callable,
        ul_objective: Callable,
        ll_model: Module,
        ul_model: Module,
        ll_var: List,
        ul_var: List,
        solver_config: Dict,
    ):
        super(CG, self).__init__(
            ll_objective,
            ul_objective,
            ul_model,
            ll_model,
            ll_var,
            ul_var,
            solver_config,
        )

        self.dynamic_initialization = "DI" in solver_config["dynamic_op"]
        self.ll_lr = solver_config["lower_level_opt"].defaults["lr"]
        self.tolerance = solver_config["CG"]["tolerance"]
        self.K = solver_config["CG"]["k"]
        self.alpha = solver_config["GDA"]["alpha_init"]
        self.alpha_decay = solver_config["GDA"]["alpha_decay"]
        self.gda_loss = solver_config.get("gda_loss", None)

    def compute_gradients(
        self,
        ll_feed_dict: Dict,
        ul_feed_dict: Dict,
        auxiliary_model: _MonkeyPatchBase,
        max_loss_iter: int = 0,
        hyper_gradient_finished: bool = False,
        next_operation: str = None,
        **kwargs
    ):
        """
        Compute the hyper-gradients of the upper-level variables with the data from feed_dict and patched models.

        Parameters
        ----------
        ll_feed_dict : Dict
            Dictionary containing the lower-level data used for optimization.
            It typically includes training data, targets, and other information required to compute the LL objective.

        ul_feed_dict : Dict
            Dictionary containing the upper-level data used for optimization.
            It typically includes validation data, targets, and other information required to compute the UL objective.

        auxiliary_model : _MonkeyPatchBase
            A patched lower model wrapped by the `higher` library.
            It serves as the lower-level model for optimization.

        max_loss_iter : int, optional
            The number of iterations used for backpropagation. Default is 0.

        hyper_gradient_finished : bool, optional
            A flag indicating whether the hyper-gradient computation is finished. Default is False.

        next_operation : str, optional
            The next operator for the calculation of the hypergradient. Default is None.

        **kwargs : dict
            Additional arguments, such as:
            - `lower_model_params` (list): Parameters of the lower-level model (default: `list(auxiliary_model.parameters())`).
            - `hparams` (list): Hyper-parameters of the upper-level model (default: `list(self.ul_var)`).

        Returns
        -------
        dict
            A dictionary containing:
            - "upper_loss": The current upper-level objective value.
            - "hyper_gradient_finished": A boolean indicating that the hyper-gradient computation is complete.

        Returns
        -------
        Dict
            A dictionary containing the upper-level objective and the status of hypergradient computation.
        """

        assert (
            not hyper_gradient_finished
        ), "CG does not support multiple hypergradient computation"
        lower_model_params = kwargs.get(
            "lower_model_params", list(auxiliary_model.parameters())
        )
        hparams = kwargs.get("hparams", list(self.ul_var))

        def fp_map(params, loss_f):
            lower_grads = list(jit.grad(loss_f, params))
            updated_params = []
            for i in range(len(params)):
                updated_params.append(params[i] - self.ll_lr * lower_grads[i])
            return updated_params

        if self.gda_loss is not None:
            ll_feed_dict["alpha"] = self.alpha * self.alpha_decay**max_loss_iter
            lower_loss = self.gda_loss(
                ll_feed_dict,
                ul_feed_dict,
                self.ul_model,
                auxiliary_model,
                params=lower_model_params,
            )
        else:
            lower_loss = self.ll_objective(
                ll_feed_dict, self.ul_model, auxiliary_model, params=lower_model_params
            )
        upper_loss = self.ul_objective(
            ul_feed_dict, self.ul_model, auxiliary_model, params=lower_model_params
        )
        if self.dynamic_initialization:
            grads_lower = jit.grad(
                upper_loss, list(auxiliary_model.parameters(time=0)), retain_graph=True
            )
            update_tensor_grads(self.ll_var, grads_lower)
        upper_grads = conjugate_gradient(
            lower_model_params,
            hparams,
            upper_loss,
            lower_loss,
            self.K,
            fp_map,
            self.tolerance,
        )

        update_tensor_grads(self.ul_var, upper_grads)

        return {"upper_loss": upper_loss.item(), "hyper_gradient_finished": True}



import abc
from typing import List, Dict
from boat_jit.utils import DynamicalSystemRules, ResultStore

importlib = __import__("importlib")

from boat_jit.operation_registry import register_class


@register_class
class DynamicalSystem(object):
    def __init__(
        self, ll_objective, ul_objective, lower_loop, ul_model, ll_model, solver_config
    ) -> None:
        """
        Abstract class for defining lower-level optimization procedures in Bilevel Optimization (BLO).

        Parameters
        ----------
        ll_objective : Callable
            The lower-level objective function of the BLO problem.
        ul_objective : Callable
            The upper-level objective function of the BLO problem.
        ll_model : jittor.Module
            The lower-level model of the BLO problem.
        ul_model : jittor.Module
            The upper-level model of the BLO problem.
        lower_loop : int
            The number of iterations for lower-level optimization.
        solver_config : Dict[str, Any]
            A dictionary containing solver configurations. It includes details about optimization algorithms, hyperparameter settings, and additional configurations required for solving the BLO problem.
        """

        self.ll_objective = ll_objective
        self.ul_objective = ul_objective
        self.lower_loop = lower_loop
        self.ul_model = ul_model
        self.ll_model = ll_model
        self.solver_config = solver_config

    @abc.abstractmethod
    def optimize(self, **kwargs):
        pass
